# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
import typing
from collections import OrderedDict
from collections.abc import Generator, Hashable, Iterable, Mapping
from typing import Literal

import numpy as np
import pandas as pd
import xarray as xr
from odc.geo.geom import Geometry, intersects
from odc.geo.gridspec import GridSpec
from typing_extensions import override

from datacube.model import Dataset, QueryField
from datacube.utils import DatacubeException

from .core import Datacube
from .query import GroupBy, Query, query_group_by

_LOG: logging.Logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from odc.geo.geobox import GeoBox

    from datacube.index import Index
    from datacube.model import Product


class GridWorkflowException(DatacubeException):
    """An ODC Exception raised while building or running Grid Workflows"""


def _fast_slice(array: xr.DataArray, indexers):
    data = array.values[indexers]
    dims = [
        dim for dim, indexer in zip(array.dims, indexers) if isinstance(indexer, slice)
    ]
    coords = OrderedDict(
        (
            dim,
            xr.Variable(
                (dim,),
                array.coords[dim].values[indexer],
                attrs=array.coords[dim].attrs,
                fastpath=True,
            ),
        )
        for dim, indexer in zip(array.dims, indexers)
        if isinstance(indexer, slice)
    )
    return xr.DataArray(data, dims=dims, coords=coords, attrs=array.attrs)


class Tile:
    """
    The Tile object holds a lightweight representation of a datacube result.

    It is produced by :meth:`.GridWorkflow.list_cells` or :meth:`.GridWorkflow.list_tiles`.

    The Tile object can be passed to :meth:`GridWorkflow.load` to be loaded into memory as
    an :class:`xr.Dataset`.

    A portion of a tile can be created by using index notation. eg:

        tile[0:1, 0:1000, 0:1000]

    This can be used to load small portions of data into memory, instead of having to access
    the entire `Tile` at once.
    """

    def __init__(self, sources: xr.DataArray, geobox: GeoBox) -> None:
        """Create a Tile representing a dataset that can be loaded.

        :param sources: An array of non-spatial dimensions of the request, holding lists of
            datacube.storage.DatasetSource objects.
        :param geobox: The spatial footprint of the Tile
        """
        self.sources: xr.DataArray = sources
        self.geobox: GeoBox = geobox

    @property
    def dims(self) -> tuple[Hashable, ...]:
        """Names of the dimensions, eg ``('time', 'y', 'x')``"""
        return self.sources.dims + self.geobox.dimensions

    @property
    def shape(self) -> tuple[int, ...]:
        """Lengths of each dimension, eg ``(285, 4000, 4000)``"""
        return self.sources.shape + self.geobox.shape

    @property
    def product(self) -> Product:
        return self.sources.values[0][0].product

    def __getitem__(self, chunk):
        sources = _fast_slice(self.sources, chunk[: len(self.sources.shape)])
        geobox = self.geobox[chunk[len(self.sources.shape) :]]
        return Tile(sources, geobox)

    # TODO(csiro) Split on time range
    def split(self, dim: str, step: int = 1) -> Generator[tuple[str, Tile]]:
        """
        Splits along a non-spatial dimension into Tile objects with a length of 1 or more in the `dim` dimension.

        :param dim: Name of the non-spatial dimension to split
        :param step: step size to split
        """
        axis = self.dims.index(dim)
        indexer = [slice(None)] * len(self.dims)
        size = self.sources[dim].size
        for i in range(0, size, step):
            indexer[axis] = slice(i, min(size, i + step))
            yield self.sources[dim].values[i], self[tuple(indexer)]

    def split_by_time(
        self, freq: str = "A", time_dim: str = "time", **kwargs
    ) -> Generator[tuple[str, Tile]]:
        """
        Splits along the `time` dimension, into periods, using pandas offsets, such as:
        :
            'A': Annual
            'Q': Quarter
            'M': Month
        See: https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects

        :param freq: time series frequency
        :param time_dim: name of the time dimension
        :param kwargs: other keyword arguments passed to ``pandas.period_range``
        :return:  generator of the key string (eg '1994') and the slice of Tile
        """
        # extract first and last timestamps from the time axis, note this will
        # work with 1 element arrays as well
        start_range, end_range = self.sources[time_dim].data[[0, -1]]

        for p in pd.period_range(start=start_range, end=end_range, freq=freq, **kwargs):
            sources_slice = self.sources.loc[
                {time_dim: slice(p.start_time, p.end_time)}
            ]
            yield str(p), Tile(sources=sources_slice, geobox=self.geobox)

    @override
    def __str__(self):
        return f"Tile<sources={self.sources!r},\n\tgeobox={self.geobox!r}>"

    @override
    def __repr__(self):
        return self.__str__()


class GridWorkflow:
    """
    GridWorkflow deals with cell- and tile-based processing using a grid defining a projection and resolution.

    Use GridWorkflow to specify your desired output grid.  The methods :meth:`list_cells` and :meth:`list_tiles`
    query the index and return a dictionary of cell or tile keys, each mapping to a :class:`Tile` object.

    The :class:`.Tile` object can then be used to load the data without needing the index,
    and can be serialized for use with the `distributed` package.
    """

    def __init__(
        self,
        index: Index,
        grid_spec: GridSpec | None = None,
        product: Product | str | None = None,
    ):
        """
        Create a grid workflow tool.

        Either grid_spec or product must be supplied.

        :param index: The database index to use.
        :param grid_spec: The grid projection and resolution
        :param product: The name of an existing product, if no grid_spec is supplied.
        """
        self.index = index

        # If available, use the provided grid_spec
        if grid_spec is not None:
            self.grid_spec: GridSpec = grid_spec
        else:
            # Otherwise, attempt to get the grid_spec by the provided product
            # which may or may not have one.
            if product is None:
                raise GridWorkflowException(
                    "Have to supply either grid_spec or product"
                )

            if isinstance(product, str):
                product = self.index.products.get_by_name(product)

            if product is None:
                raise GridWorkflowException("No such product", product)

            if product.grid_spec is None:
                raise GridWorkflowException(
                    "Supplied product does not have a gridspec", product
                )

            self.grid_spec = product.grid_spec

    def cell_observations(
        self,
        cell_index: tuple[int, int] | None = None,
        geopolygon: Geometry | None = None,
        tile_buffer: tuple[float, float] | None = None,
        **indexers: QueryField,
    ) -> dict[tuple[int, int], dict[str, Dataset | GeoBox]]:
        """
        List datasets, grouped by cell.

        :param geopolygon:
            Only return observations with data inside polygon.
        :param tile_buffer:
            buffer tiles by (y, x) in CRS units
        :param  cell_index:
            The cell index. E.g. (14, -40)
        :param indexers:
            Query to match the datasets, see :py:class:`datacube.api.query.Query`
        :return: A dictionary of cell index (int, int) mapping to a dict containing two keys,
          "datasets", with a list of datasets, and "geobox", containing the geobox for the cell.

        .. seealso::
            :meth:`datacube.Datacube.find_datasets`

            :class:`datacube.api.query.Query`
        """
        # pylint: disable=too-many-locals
        # TODO: split this method into 3: cell/polygon/unconstrained querying

        if tile_buffer is not None and geopolygon is not None:
            raise GridWorkflowException(
                "Cannot process tile_buffering and geopolygon together."
            )
        cells: dict[tuple[int, int], dict[str, Dataset | GeoBox]] = {}

        def add_dataset_to_cells(tile_index, tile_geobox, dataset_):
            cells.setdefault(tile_index, {"datasets": [], "geobox": tile_geobox})[
                "datasets"
            ].append(dataset_)

        if cell_index:
            geobox = self.grid_spec.tile_geobox(cell_index)
            geobox = geobox.buffered(*tile_buffer) if tile_buffer else geobox

            datasets, query = self._find_datasets(geobox.extent, indexers)
            for dataset in datasets:
                if intersects(geobox.extent, dataset.extent.to_crs(self.grid_spec.crs)):
                    add_dataset_to_cells(cell_index, geobox, dataset)
            return cells
        else:
            datasets, query = self._find_datasets(geopolygon, indexers)
            geobox_cache: dict[tuple[int, int], GeoBox] = {}

            if query.geopolygon:
                # Get a rough region of tiles
                query_tiles = {
                    tile_index
                    for tile_index, tile_geobox in self.grid_spec.tiles_from_geopolygon(
                        query.geopolygon, geobox_cache=geobox_cache
                    )
                }

                for dataset in datasets:
                    # Go through our datasets and see which tiles each dataset produces, and whether they intersect
                    # our query geopolygon.
                    dataset_extent = dataset.extent.to_crs(self.grid_spec.crs)
                    bbox = dataset_extent.boundingbox
                    bbox = bbox.buffered(*tile_buffer) if tile_buffer else bbox

                    for tile_index, tile_geobox in self.grid_spec.tiles(
                        bbox, geobox_cache=geobox_cache
                    ):
                        if tile_index in query_tiles and intersects(
                            tile_geobox.extent, dataset_extent
                        ):
                            add_dataset_to_cells(tile_index, tile_geobox, dataset)

            else:
                for dataset in datasets:
                    dataset_extent = (
                        dataset.extent.buffer(*tile_buffer)
                        if tile_buffer
                        else dataset.extent
                    )
                    for tile_index, tile_geobox in self.grid_spec.tiles_from_geopolygon(
                        dataset_extent, geobox_cache=geobox_cache
                    ):
                        if tile_buffer:
                            tile_geobox = tile_geobox.buffered(*tile_buffer)
                        add_dataset_to_cells(tile_index, tile_geobox, dataset)

            return cells

    def _find_datasets(self, geopolygon, indexers):
        query = Query(index=self.index, geopolygon=geopolygon, **indexers)
        if not query.product:
            raise RuntimeError("must specify a product")
        datasets = self.index.datasets.search_eager(**query.search_terms)
        return datasets, query

    @staticmethod
    def group_into_cells(
        observations, group_by: GroupBy
    ) -> dict[tuple[int, int], Tile]:
        """
        Group observations into a stack of source tiles.

        :param observations: datasets grouped by cell index, like from :py:meth:`cell_observations`
        :param group_by: grouping method, as returned by :py:meth:`datacube.api.query.query_group_by`
        :return: tiles grouped by cell index

        .. seealso::
            :meth:`load`

            :meth:`datacube.Datacube.group_datasets`
        """
        cells = {}
        for cell_index, observation in observations.items():
            sources = Datacube.group_datasets(observation["datasets"], group_by)
            cells[cell_index] = Tile(sources, observation["geobox"])
        return cells

    @staticmethod
    def tile_sources(
        observations, group_by: GroupBy
    ) -> dict[tuple[int, int, np.datetime64], Tile]:
        """
        Split observations into tiles and group into source tiles

        :param observations: datasets grouped by cell index, like from :meth:`cell_observations`
        :param group_by: grouping method, as returned by :py:meth:`datacube.api.query.query_group_by`
        :return: tiles grouped by cell index and time

        .. seealso::
            :meth:`load`

            :meth:`datacube.Datacube.group_datasets`
        """
        tiles = {}
        for cell_index, observation in observations.items():
            dss = observation["datasets"]
            geobox = observation["geobox"]

            sources = Datacube.group_datasets(dss, group_by)
            coord = sources[sources.dims[0]]
            for i in range(coord.size):
                tile_index = cell_index + (coord.values[i],)
                tiles[tile_index] = Tile(sources[i : i + 1], geobox)

        return tiles

    def list_cells(
        self, cell_index: tuple[int, int] | None = None, **query
    ) -> dict[tuple[int, int], Tile]:
        """
        List cells that match the query.

        Returns a dictionary of cell indexes to :class:`.Tile` objects.

        Cells are included if they contain any datasets that match the query using the same format as
        :meth:`datacube.Datacube.load`.

        E.g.::

            gw.list_cells(product='ls5_nbar_albers',
                          time=('2001-1-1 00:00:00', '2001-3-31 23:59:59'))

        :param cell_index: The cell index. E.g. (14, -40)
        :param query: see :py:class:`datacube.api.query.Query`
        """
        observations = self.cell_observations(cell_index, **query)
        return self.group_into_cells(observations, query_group_by(**query))

    def list_tiles(
        self, cell_index: tuple[int, int] | None = None, **query
    ) -> dict[tuple[int, int, np.datetime64], Tile]:
        """
        List tiles of data, sorted by cell.
        ::

            tiles = gw.list_tiles(product='ls5_nbar_albers',
                                  time=('2001-1-1 00:00:00', '2001-3-31 23:59:59'))

        The values can be passed to :meth:`load`

        :param cell_index: The cell index (optional). E.g. (14, -40)
        :param query: see :py:class:`datacube.api.query.Query`

        .. seealso:: :meth:`load`
        """
        observations = self.cell_observations(cell_index, **query)
        return self.tile_sources(observations, query_group_by(**query))

    @staticmethod
    def load(
        tile: Tile,
        measurements: Iterable[str] | None = None,
        dask_chunks: Mapping[str, int | Literal["auto"]] | None = None,
        fuse_func=None,
        resampling: str | dict | None = None,
        skip_broken_datasets: bool = False,
    ) -> xr.Dataset:
        """
        Load data for a cell/tile.

        The data to be loaded is defined by the output of :meth:`list_tiles`.

        This is a static function and does not use the index. This can be useful
        to minimize the number of database connections when running as a worker
        in a distributed environment.

        See the documentation on using `xr with dask <http://xr.pydata.org/en/stable/dask.html>`_
        for more information.

        :param tile: The tile to load.

        :param measurements: The names of measurements to load

        :param dask_chunks: If the data should be loaded as needed using :py:class:`dask.array.Array`,
            specify the chunk size in each output direction.

            See the documentation on using `xr with dask <http://xr.pydata.org/en/stable/dask.html>`_
            for more information.

        :param fuse_func: Function to fuse together a tile that has been pre-grouped by calling
            :meth:`list_cells` with a ``group_by`` parameter.

        :param resampling:

            The resampling method to use if re-projection is required, could be
            configured per band using a dictionary (:meth: `load_data`)

            Valid values are: ``'nearest', 'cubic', 'bilinear', 'cubic_spline', 'lanczos', 'average'``

            Defaults to ``'nearest'``.

        :param skip_broken_datasets: If True, ignore broken datasets and continue processing with the data
             that can be loaded. If False, an exception will be raised on a broken dataset. Defaults to False.


        .. seealso::
            :meth:`list_tiles` :meth:`list_cells`
        """
        measurement_dicts = tile.product.lookup_measurements(measurements)

        dataset = Datacube.load_data(
            tile.sources,
            tile.geobox,
            measurement_dicts,
            resampling=resampling,
            dask_chunks=dask_chunks,
            fuse_func=fuse_func,
            skip_broken_datasets=skip_broken_datasets,
        )

        return dataset

    def update_tile_lineage(self, tile: Tile):
        for i in range(tile.sources.size):
            sources = tile.sources.values[i]
            tile.sources.values[i] = tuple(
                self.index.datasets.get(dataset.id, include_sources=True)
                for dataset in sources
            )
        return tile

    @override
    def __str__(self):
        return f"GridWorkflow<index={self.index!r},\n\tgridspec={self.grid_spec!r}>"

    @override
    def __repr__(self):
        return self.__str__()



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import contextlib
import logging
from collections.abc import Mapping
from pathlib import Path
from typing import Any

import netCDF4
import xarray
from odc.geo import CRS
from xarray.core.coordinates import DatasetCoordinates

from datacube.storage._hdf5 import HDF5_LOCK
from datacube.utils import DatacubeException

from . import writer as netcdf_writer

_LOG: logging.Logger = logging.getLogger(__name__)


def _get_units(coord) -> str:
    """
    Guess units from coordinate
    1. Value of .units if set
    2. 'seconds since 1970-01-01 00:00:00' if coord dtype is datetime
    3. '1' otherwise
    """
    units = getattr(coord, "units", None)
    if units is not None:
        return units
    dtype = getattr(coord.values, "dtype", None)
    if dtype is None:
        return "1"

    if dtype.kind == "M":
        return "seconds since 1970-01-01 00:00:00"

    return "1"


def create_netcdf_storage_unit(
    filename: Path,
    crs: CRS,
    coordinates: DatasetCoordinates,
    variables: Mapping[str, Any],
    variable_params: Mapping[str, Mapping[str, Any]],
    global_attributes: dict | None = None,
    netcdfparams: dict | None = None,
) -> netCDF4.Dataset:
    """
    Create a NetCDF file on disk.

    :param filename: filename to write to
    :param crs: Datacube CRS object defining the spatial projection
    :param coordinates: Dict of named `datacube.model.Coordinate`s to create
    :param variables: Dict of named `datacube.model.Variable`s to create
    :param variable_params:
        Dict of dicts, with keys matching variable names, of extra parameters for variables
    :param global_attributes: named global attributes to add to output file
    :param netcdfparams: Extra parameters to use when creating netcdf file
    :return: open netCDF4.Dataset object, ready for writing to
    """
    filename = Path(filename)
    if filename.exists():
        raise RuntimeError(f"Storage Unit already exists: {filename}")

    with contextlib.suppress(OSError):
        filename.parent.mkdir(parents=True)

    _LOG.info("Creating storage unit: %s", filename)

    nco = netcdf_writer.create_netcdf(str(filename), **(netcdfparams or {}))

    for name, coord in coordinates.items():
        if coord.values.ndim > 0:  # skip CRS coordinate
            netcdf_writer.create_coordinate(
                nco, str(name), coord.values, _get_units(coord)
            )

    grid_mapping = netcdf_writer.DEFAULT_GRID_MAPPING
    netcdf_writer.create_grid_mapping_variable(nco, crs, name=grid_mapping)

    for name, variable in variables.items():
        has_crs = all(dim in variable.dims for dim in crs.dimensions)
        var_params = variable_params.get(name, {})
        data_var = netcdf_writer.create_variable(
            nco,
            name,
            variable,
            grid_mapping=grid_mapping if has_crs else None,
            **var_params,
        )

        for key, value in var_params.get("attrs", {}).items():
            setattr(data_var, key, value)

    for key, value in (global_attributes or {}).items():
        setattr(nco, key, value)

    return nco


def write_dataset_to_netcdf(
    dataset: xarray.Dataset,
    filename: Path,
    global_attributes: dict | None = None,
    variable_params: dict | None = None,
    netcdfparams=None,
) -> None:
    """
    Write a Data Cube style xarray Dataset to a NetCDF file

    Requires a spatial Dataset, with attached coordinates and global crs attribute.

    :param dataset:
    :param filename: Output filename
    :param global_attributes: Global file attributes. dict of attr_name: attr_value
    :param variable_params: dict of variable_name: {param_name: param_value, [...]}
                            Allows setting storage and compression options per variable.
                            See the `netCDF4.Dataset.createVariable` for available
                            parameters.
    :param netcdfparams: Optional params affecting netCDF file creation
    """
    global_attributes = global_attributes or {}
    variable_params = variable_params or {}
    filename = Path(filename)

    if not dataset.data_vars.keys():
        raise DatacubeException("Cannot save empty dataset to disk.")

    if dataset.odc.geobox is None:
        raise DatacubeException(
            "Dataset geobox property is None, cannot write to NetCDF file."
        )

    try:
        HDF5_LOCK.acquire(blocking=True)
        nco = create_netcdf_storage_unit(
            filename,
            dataset.odc.geobox.crs,
            dataset.coords,
            dataset.data_vars,
            variable_params,
            global_attributes,
            netcdfparams,
        )

        for name, variable in dataset.data_vars.items():
            nco[name][:] = netcdf_writer.netcdfy_data(variable.values)

        nco.close()
    finally:
        HDF5_LOCK.release()



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""Reader driver construction for tests"""

from pathlib import Path

from datacube.drivers._types import ReaderDriver
from datacube.drivers.rio._reader import (
    RDEntry,
)
from datacube.storage import BandInfo
from datacube.testutils import mk_sample_dataset
from datacube.testutils.threads import FakeThreadPoolExecutor

NetCDF = "NetCDF"  # pylint: disable=invalid-name
GeoTIFF = "GeoTIFF"  # pylint: disable=invalid-name


def mk_rio_driver() -> ReaderDriver:
    pool = FakeThreadPoolExecutor()
    rde = RDEntry()
    return rde.new_instance({"pool": pool, "allow_custom_pool": True})


def mk_band(
    name: str,
    base_uri: str,
    path: str = "",
    format: str = GeoTIFF,  # pylint: disable=redefined-builtin  # noqa: A002
    **extras,
) -> BandInfo:
    """
    **extras**:
       layer, band, nodata, dtype, units, aliases
    """
    band_opts = {
        k: extras.pop(k)
        for k in ["path", "layer", "band", "nodata", "dtype", "units", "aliases"]
        if k in extras
    }

    band = dict(name=name, path=path, **band_opts)
    ds = mk_sample_dataset([band], base_uri, format=format, **extras)
    return BandInfo(ds, name)


def open_reader(
    path: str,
    band_name: str = "b1",
    format: str = GeoTIFF,  # pylint: disable=redefined-builtin  # noqa: A002
    **extras,
):
    """
    **extras**:
       layer, band, nodata, dtype, units, aliases
    """
    rdr = mk_rio_driver()
    base_uri = Path(path).absolute().as_uri()
    bi = mk_band(band_name, base_uri, format=format, **extras)
    load_ctx = rdr.new_load_context(iter([bi]), None)
    fut = rdr.open(bi, load_ctx)
    return fut.result()


def tee_new_load_context(rdr, new_impl) -> None:
    """When calling rdr.new_load_context(bands, old_ctx) tee data to new_impl"""
    _real_impl = rdr.new_load_context

    def patched(bands, old_ctx):
        bands = list(bands)
        new_impl(iter(bands), old_ctx)
        return _real_impl(iter(bands), old_ctx)

    rdr.new_load_context = patched



import sys
from datetime import datetime, timezone
from time import monotonic

from odc.geo.geom import CRS, polygon

from datacube import Datacube
from datacube.model import Range


def benchmark(test, dc, label, n):
    total = 0.0
    total_first = 0.0
    last_count = None
    for i in range(n):
        start = monotonic()
        # count, first = test(dc)
        count = 0
        first = None
        for _ in test(dc):
            if not count:
                first = monotonic()
            count += 1
        if count == 0:
            first = start
        end = monotonic()
        if last_count and count != last_count:
            print(f"Count mismatch in {label}: {count} vs {last_count}")
        last_count = count
        print(
            f"Test {label}#{i + 1}: {end - start}s   ({first - start}s to first returned dataset)"
        )
        total += end - start
        total_first += first - start
    print(f"Test {label}-count: {count} rows")
    print(f"Test {label}-avg: {total / n}s  ({total / (n * count)})s/row")
    print(
        f"Test {label}-avg-to-first-return: {total_first / n}s  ({total / (n * count)})s/row"
    )
    print()
    print("-----------------------------------------------------------------")


def test_less_than(dc):
    return dc.index.datasets.search(
        product="ga_ls8c_ard_3", cloud_cover=Range(None, 0.2)
    )


def test_geospatial_search(dc):
    return dc.index.datasets.search(
        product="ga_ls8c_ard_3",
        lat=Range(-30.0, -25.0),
        lon=Range(140.0, 145.0),
    )


def test_offset_geom(dc):
    if dc.index.supports_external_lineage:
        return dc.index.datasets.search(
            product="ga_ls8c_ard_3",
            geometry=polygon(
                [
                    [140.0, -25.0],
                    [142.0, -25.0],
                    [145.0, -30.0],
                    [145.0, -30.0],
                    [140.0, -25.0],
                ],
                crs=CRS("epsg:4326"),
            ),
        )
    else:
        return dc.index.datasets.search(
            product="ga_ls8c_ard_3",
            lat=Range(-30.0, -25.0),
            lon=Range(140.0, 145.0),
        )


def test_temporal_search(dc):
    return dc.index.datasets.search(
        product="ga_ls8c_ard_3",
        time=Range(
            datetime(2016, 1, 1, tzinfo=timezone.utc),
            datetime(2016, 4, 5, tzinfo=timezone.utc),
        ),
    )


def main(args):
    env = args.pop() if args else "datacube_real"
    print("Testing on database ", env)
    dc = Datacube(env=env)
    benchmark(test_less_than, dc, "less_than", 20)
    benchmark(test_geospatial_search, dc, "geospatial", 20)
    benchmark(test_temporal_search, dc, "temporal", 20)
    benchmark(test_offset_geom, dc, "geom", 20)


if __name__ == "__main__":
    args = sys.argv[1:]
    main(args)

# For custom CRS search tests (TODO)
# BoundingBox(
#       left=762759.2567816022, bottom=-3326371.8490792206,
#       right=1295116.9248742603, top=-2727561.09954842, crs=CRS('EPSG:3577'))



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Tracking spatial indexes
"""

import logging
from collections.abc import Mapping
from threading import Lock

from antimeridian import fix_shape
from geoalchemy2 import Geometry
from odc.geo import CRS
from odc.geo import Geometry as Geom
from odc.geo.geom import multipolygon, polygon
from sqlalchemy import ForeignKey, delete, select, text
from sqlalchemy.dialects import postgresql as postgres
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session, mapped_column
from sqlalchemy.sql.ddl import DropTable

from ._core import METADATA
from ._schema import Base, Dataset, SpatialIndex, SpatialIndexRecord, orm_registry
from .sql import SCHEMA_NAME

_LOG: logging.Logger = logging.getLogger(__name__)


# In theory, we could just use the SQLAlchemy registry for this, but it is not indexed
# in a useful way.
class SpatialIndexORMRegistry:
    """Threadsafe global registry of SpatialIndex ORM classes, indexed by EPSG/SRID code."""

    _registry: dict[int, type[SpatialIndex]] = {}
    _lock = Lock()

    def __init__(self) -> None:
        self._registry = self.__class__._registry
        self._lock = self.__class__._lock

    def _to_epsg(self, epsg_or_crs: CRS | int) -> int:
        """Utility method to convert an epsg_or_crs to an epsg."""
        if isinstance(epsg_or_crs, CRS):
            if epsg_or_crs.epsg is None:
                raise ValueError("CRS with no epsg number")
            return epsg_or_crs.epsg
        else:
            return epsg_or_crs

    def register(self, epsg_or_crs: CRS | int) -> bool:
        """Ensure that SpatialIndex ORM clss is registered for this EPSG/SRID"""
        epsg = self._to_epsg(epsg_or_crs)
        added = False
        with self._lock:
            if epsg not in self._registry:
                self._registry[epsg] = self._mint_new_spindex(epsg)
                added = True
        return added

    def get(self, epsg_or_crs: CRS | int) -> type[SpatialIndex] | None:
        """Retrieve the registered SpatialIndex ORM class"""
        epsg = self._to_epsg(epsg_or_crs)
        return self._registry.get(epsg)

    def _mint_new_spindex(self, epsg: int) -> type[SpatialIndex]:
        """
        Dynamically create a new ORM class for a EPSG/SRID.

        Note: Called within registry lock.
        """
        attributes = {
            "__tablename__": f"spatial_{epsg}",
            "__table_args__": (
                METADATA,
                {
                    "schema": SCHEMA_NAME,
                    "comment": "A product or dataset type, family of related datasets.",
                },
            ),
            "dataset_ref": mapped_column(
                postgres.UUID(as_uuid=True),
                ForeignKey(Dataset.id),
                primary_key=True,
                nullable=False,
                comment="The dataset being indexed",
            ),
            "extent": mapped_column(
                Geometry("MULTIPOLYGON", srid=epsg),
                nullable=False,
                comment="The extent of the dataset",
            ),
        }
        return type(f"SpatialIdx{epsg}", (SpatialIndex, Base), attributes)


def is_spindex_table_name(name: str) -> bool:
    bits = name.split("_")
    if len(bits) == 2 and bits[0] == "spatial":
        try:
            srid = int(bits[1])
            return srid > 0
        except ValueError:
            pass
    return False


def spindex_for_epsg(epsg: int) -> type[SpatialIndex]:
    """Return ORM class of a SpatialIndex for EPSG/SRID - dynamically creating if necessary"""
    sir = SpatialIndexORMRegistry()
    spindex = sir.get(epsg)
    if spindex is None:
        sir.register(epsg)
        spindex = sir.get(epsg)
        assert spindex is not None  # for type-checker
    return spindex


def crs_to_epsg(crs: CRS) -> int:
    if not str(crs).upper().startswith("EPSG:") and crs.epsg is None:
        raise ValueError("Non-EPSG-style CRS.")
    elif crs.epsg is not None:
        return crs.epsg
    else:
        return int(str(crs)[5:])


def spindex_for_crs(crs: CRS) -> type[SpatialIndex]:
    """Return ORM class of a SpatialIndex for CRS - dynamically creating if necessary"""
    try:
        return spindex_for_epsg(crs_to_epsg(crs))
    except ValueError:
        # Postgis identifies CRSs by a numeric "SRID" which is equivalent to EPSG number.
        raise ValueError(
            f"Cannot create a postgis spatial index for a non-EPSG-style CRS: {crs!s}"
        ) from None


def spindex_for_record(rec: SpatialIndexRecord) -> type[SpatialIndex]:
    """Convert a Record of a SpatialIndex created in a particular database to an ORM class"""
    return spindex_for_epsg(rec.srid)


def ensure_spindex(engine: Engine, sp_idx: type[SpatialIndex], crs_id: int) -> None:
    """Ensure a Spatial Index exists in a particular database."""
    with Session(engine) as session:
        results = session.execute(
            select(SpatialIndexRecord.srid).where(SpatialIndexRecord.srid == crs_id)
        )
        for _ in results:
            # SpatialIndexRecord exists - actual index assumed to exist too.
            return
        # SpatialIndexRecord doesn't exist - create the index table...
        orm_registry.metadata.create_all(engine, [sp_idx.__table__])  # type: ignore[attr-defined]
        # ... and add a SpatialIndexRecord
        session.add(SpatialIndexRecord(srid=crs_id, table_name=sp_idx.__tablename__))  # type: ignore[attr-defined]
        session.commit()
        session.flush()
        # Permissions Management
        for command in [
            # Read access to odc_user
            f"grant select on {SCHEMA_NAME}.{sp_idx.__tablename__} to odc_user;",  # type: ignore[attr-defined]
            # Insert access to odc_manage
            f"grant insert on {SCHEMA_NAME}.{sp_idx.__tablename__} to odc_manage;",  # type: ignore[attr-defined]
            # Full access to odc_admin
            f"grant all on {SCHEMA_NAME}.{sp_idx.__tablename__} to odc_admin;",  # type: ignore[attr-defined]
        ]:
            session.execute(text(command))
        # Grant statements in PostgreSQL behave like transactions, so commit them.
        session.commit()


def drop_spindex(engine: Engine, sp_idx: type[SpatialIndex], crs_id: int) -> bool:
    with Session(engine) as session:
        results = session.execute(
            select(SpatialIndexRecord).where(SpatialIndexRecord.srid == crs_id)
        )
        spidx_record = None
        for result in results:
            spidx_record = result[0]
            break
        record_del_result = False
        if spidx_record:
            del_res = session.execute(
                delete(SpatialIndexRecord).where(
                    SpatialIndexRecord.srid == spidx_record.srid
                )
            )
            record_del_result = del_res.rowcount == 1

        drop_res = session.execute(
            DropTable(sp_idx.__table__, if_exists=True)  # type: ignore[attr-defined]
        )
        drop_table_result = drop_res.rowcount == 1  # type: ignore[attr-defined]
        _LOG.warning(
            f"spindex record deleted: {record_del_result}   table dropped: {drop_table_result}"
        )

    return True


def spindexes(engine: Engine) -> Mapping[int, type[SpatialIndex]]:
    """
    Return a SRID-to-Spatial Index ORM class mapping for indexes that exist in a particular database.
    """
    out = {}
    with Session(engine) as session:
        results = session.execute(select(SpatialIndexRecord.srid))
        for result in results:
            epsg = int(result[0])
            spindex = spindex_for_epsg(epsg)
            out[epsg] = spindex
    return out


def promote_to_multipolygon(geom: Geom) -> Geom:
    # Assumes input is a polygon or multipolygon - does not work on lines or points
    if geom.geom_type == "MultiPolygon":
        return geom
    elif geom.geom_type == "Polygon":
        # Promote to multipolygon (is there a more elegant way to do this??)
        polycoords = [list(geom.geom.exterior.coords)]
        for interior in geom.geom.interiors:
            polycoords.append(list(interior.coords))
        geom = multipolygon([polycoords], crs=geom.crs)
        return geom
    else:
        raise ValueError(
            f"Cannot promote geometry type {geom.geom_type} to multi-polygon"
        )


def geom_alchemy(geom: Geom) -> str:
    geom = promote_to_multipolygon(geom)
    if geom.crs is None:
        raise ValueError("Geometry with no CRS")
    epsg = crs_to_epsg(geom.crs)
    return f"SRID={epsg};{geom.wkt}"


# 4326-like EPSG codes.
# extents in these CRSs will be projected to 4326, antimeridian-fixed, and then projected back, so it
# is very important that only CRSs where this is appropriate are included.
# It should only include CRSs that use the WGS84 datum, an equatorial cylindrical projection and are centred on
# the prime meridian (and therefore have a discontinuity at the anti-meridian).
#
# Just epsg:3857 for now (Web Mercator)
EPSG4326_LIKE_CODES = [3857]


def sanitise_extent(extent, crs) -> Geom:
    if crs.epsg == 4326:
        prelim = extent.to_crs(crs)
        return Geom(fix_shape(prelim.geom), crs=crs)
    elif crs.epsg in EPSG4326_LIKE_CODES:
        prelim = extent.to_crs("epsg:4326")
        fixed = Geom(fix_shape(prelim.geom), crs="epsg:4326")
        return fixed.to_crs(crs)
    else:
        return extent.to_crs(crs)


def generate_dataset_spatial_values(
    dataset_id, crs, extent: Geom | None
) -> dict[str, str] | None:
    extent = sanitise_extent(extent, crs)
    if extent is None:
        return None
    geom_alch = geom_alchemy(extent)
    return {"dataset_ref": dataset_id, "extent": geom_alch}


def extract_geometry_from_eo3_projection(eo3_gs_doc) -> Geom | None:
    native_crs = CRS(eo3_gs_doc["spatial_reference"])
    valid_data = eo3_gs_doc.get("valid_data")
    if valid_data:
        return Geom(valid_data, crs=native_crs)
    else:
        geo_ref_points = eo3_gs_doc.get("geo_ref_points")
        if geo_ref_points:
            return polygon(
                [
                    (geo_ref_points[key]["x"], geo_ref_points[key]["y"])
                    for key in ("ll", "ul", "ur", "lr", "ll")
                ],
                crs=native_crs,
            )
        else:
            return None



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import datetime
import decimal
from textwrap import dedent

import pytest
import yaml

from datacube.model import Range, metadata_from_doc
from datacube.model.fields import Expression, get_dataset_fields, parse_search_field

METADATA_DOC = yaml.safe_load("""---
name: test
description: test all simple search field types
dataset:
  id: [id]
  sources: [lineage, source_datasets]
  label: [label]
  creation_dt: [creation_dt]
  search_fields:
    x_default_type:
       description: string type is assumed
       offset: [some, path, x_default_type_path]

    x_string:
      type: string
      description: field of type 'string'
      offset: [x_string_path]

    x_double:
      type: double
      description: field of type 'double'
      offset: [x_double_path]

    x_integer:
      type: integer
      description: field of type 'integer'
      offset: [x_integer_path]

    x_numeric:
      type: numeric
      description: field of type 'numeric'
      offset: [x_numeric_path]

    x_datetime:
      type: datetime
      description: field of type 'datetime'
      offset: [x_datetime_path]
""")

SAMPLE_DOC = yaml.safe_load("""---
id: usually_a_uuid_str
x_string_path: some_string
x_double_path: 6.283185307179586
x_integer_path: 4466778
x_numeric_path: '100.33'
x_datetime_path: 1999-04-15 12:33:55.001
some:
  path:
    x_default_type_path: just_a_string
""")

METADATA_DOC_RANGES = yaml.safe_load("""---
name: test
description: test all simple search field types
dataset:
  id: [id]
  sources: [lineage, source_datasets]
  label: [label]
  creation_dt: [creation_dt]
  search_fields:
     t_range:
       type: datetime-range
       min_offset: [[t,a], [t,b]]
       max_offset: [[t,a], [t,b]]

     x_range:
       type: double-range
       min_offset: [[x,a], [x,b], [x,c], [x,d]]
       max_offset: [[x,a], [x,b], [x,c], [x,d]]

     float_range:
       type: float-range
       description: float-range is alias for numeric-range
       min_offset: [[a]]
       max_offset: [[b]]

     ab:
       type: integer-range
       min_offset: [[a]]
       max_offset: [[b]]
""")

SAMPLE_DOC_RANGES = yaml.safe_load("""---
t:
  a: 1999-04-15
  b: 1999-04-16
x:
  a: 1
  b: 2
  c: 3
  d: 4
""")


def test_get_dataset_simple_fields() -> None:
    xx = get_dataset_fields(METADATA_DOC)
    assert xx["x_default_type"].type_name == "string"

    type_map = {
        "double": float,
        "integer": int,
        "string": str,
        "datetime": datetime.datetime,
        "numeric": decimal.Decimal,
    }

    for n, f in xx.items():
        assert n == f.name
        assert isinstance(f.description, str)

        expected_type = type_map.get(f.type_name)
        assert expected_type is not None
        vv = f.extract(SAMPLE_DOC)
        assert isinstance(vv, expected_type)

        # missing data should return None
        assert f.extract({}) is None


def test_get_dataset_range_fields() -> None:
    xx = get_dataset_fields(METADATA_DOC_RANGES)
    v = xx["x_range"].extract(SAMPLE_DOC_RANGES)
    assert v == Range(1, 4)

    v = xx["t_range"].extract(SAMPLE_DOC_RANGES)
    assert v.begin.strftime("%Y-%m-%d") == "1999-04-15"
    assert v.end.strftime("%Y-%m-%d") == "1999-04-16"

    # missing range should return None
    assert xx["ab"].extract({}) is None

    # partially missing Range
    assert xx["ab"].extract({"a": 3}) == Range(3, None)
    assert xx["ab"].extract({"b": 4}) == Range(None, 4)

    # float-range conversion
    assert xx["float_range"].type_name == "numeric-range"


def test_metadata_from_doc() -> None:
    mm = metadata_from_doc(METADATA_DOC)
    assert mm.definition is METADATA_DOC

    rdr = mm.dataset_reader(SAMPLE_DOC)
    assert rdr.x_double == SAMPLE_DOC["x_double_path"]
    assert rdr.x_integer == SAMPLE_DOC["x_integer_path"]
    assert rdr.x_string == SAMPLE_DOC["x_string_path"]
    assert rdr.x_numeric == decimal.Decimal(SAMPLE_DOC["x_numeric_path"])


def test_bad_field_definition() -> None:
    def doc(s):
        return yaml.safe_load(dedent(s))

    with pytest.raises(ValueError):
        parse_search_field(
            doc("""
        type: bad_type
        offset: [a]
        """)
        )

    with pytest.raises(ValueError):
        parse_search_field(
            doc("""
        type: badtype-range
        offset: [a]
        """)
        )

    with pytest.raises(ValueError):
        parse_search_field(
            doc("""
        type: double
        description: missing offset
        """)
        )

    with pytest.raises(ValueError):
        parse_search_field(
            doc("""
        type: double-range
        description: missing min_offset
        max_offset: [[a]]
        """)
        )

    with pytest.raises(ValueError):
        parse_search_field(
            doc("""
        type: double-range
        description: missing max_offset
        min_offset: [[a]]
        """)
        )


def test_expression() -> None:
    assert Expression() == Expression()
    assert (Expression() == object()) is False



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from queue import Queue

from datacube.utils.generic import (
    it2q,
    map_with_lookahead,
    qmap,
    thread_local_cache,
)


def test_map_with_lookahead() -> None:
    def if_one(x) -> str:
        return "one" + str(x)

    def if_many(x) -> str:
        return "many" + str(x)

    assert list(map_with_lookahead(iter([]), if_one, if_many)) == []
    assert list(map_with_lookahead(iter([1]), if_one, if_many)) == [if_one(1)]
    assert list(map_with_lookahead(range(5), if_one, if_many)) == list(
        map(if_many, range(5))
    )
    assert list(map_with_lookahead(range(10), if_one=if_one)) == list(range(10))
    assert list(map_with_lookahead(iter([1]), if_many=if_many)) == [1]


def test_qmap() -> None:
    q: Queue = Queue(maxsize=100)
    it2q((i for i in range(10)), q)
    rr = list(qmap(str, q))
    assert rr == [str(x) for x in range(10)]
    q.join()  # should not block


def test_thread_local_cache() -> None:
    name = "test_0123394"
    v: dict = {}

    assert thread_local_cache(name, v) is v
    assert thread_local_cache(name) is v
    assert thread_local_cache(name, purge=True) is v
    assert thread_local_cache(name, 33) == 33
    assert thread_local_cache(name, purge=True) == 33

    assert thread_local_cache("no_such_key", purge=True) is None
    assert thread_local_cache("no_such_key", 111, purge=True) == 111



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import collections.abc
import datetime
import logging
import uuid
import warnings
from collections.abc import Callable, Hashable, Iterable, Mapping, Sequence
from itertools import groupby
from typing import TYPE_CHECKING, Any, Literal, TypeAlias, cast

import deprecat
import numpy
import xarray
from dask import array as da
from odc.geo import CRS, XY, Resolution, res_, resyx_, yx_
from odc.geo.geobox import GeoBox, GeoboxTiles
from odc.geo.geom import Geometry, bbox_union, box, intersects
from odc.geo.warp import Resampling
from odc.geo.xr import xr_coords
from typing_extensions import override
from xarray.core.coordinates import DataArrayCoordinates

from datacube.cfg import GeneralisedCfg, GeneralisedEnv, GeneralisedRawCfg, ODCConfig
from datacube.model import Dataset, ExtraDimensions, ExtraDimensionSlices, Measurement
from datacube.model.utils import xr_apply
from datacube.storage import BandInfo, reproject_and_fuse
from datacube.utils import ignore_exceptions_if
from datacube.utils.dates import normalise_dt

if TYPE_CHECKING:
    from odc.geo.crs import MaybeCRS
    from pandas import DataFrame

    from datacube.model import GridSpec
    from datacube.utils.geometry import GeoBox as LegacyGeoBox

from ..drivers import new_datasource
from ..index import Index, extract_geom_from_query, index_connect
from ..migration import ODC2DeprecationWarning
from ..model import QueryField
from ..storage._load import FuserFunction, ProgressFunction
from .query import GroupBy, Query, _normalise_geobox, query_group_by

_LOG: logging.Logger = logging.getLogger(__name__)


DataFrameLike: TypeAlias = list[dict[str, str | int | float | None]]


class TerminateCurrentLoad(Exception):  # noqa: N818
    """This exception is raised by user code from `progress_cbk`
    to terminate currently running `.load`
    """

    pass


class Datacube:
    """
    Interface to search, read and write a datacube.
    """

    def __init__(
        self,
        index: Index | None = None,
        config: GeneralisedCfg | None = None,
        env: GeneralisedEnv | None = None,
        raw_config: GeneralisedRawCfg | None = None,
        app: str | None = None,
        validate_connection: bool = True,
    ) -> None:
        """
        Create an interface for the query and storage access.

        :param index: The database index to use. If provided, config, app, env and raw_config should all be None.

        :param config: One of:
            - None (Use provided ODCEnvironment or Index, or perform default config loading.)
            - An ODCConfig object
            - A file system path pointing to the location of the config file.
            - A list of file system paths to search for config files. The first readable file found will be used.
            If an index or an explicit ODCEnvironment is supplied, config and raw_config should be None.

        :param env: The datacube environment to use.
            Either an explicit ODCEnvironment object, or a str which is a section name in the loaded config file.

            Defaults to 'default'. Falls back to 'datacube' with a deprecation warning if config file does not
            contain a 'default' section.

            Allows you to have multiple datacube instances in one configuration, specified on load,
            e.g. 'dev', 'test' or 'landsat', 'modis' etc.

            If env is an ODCEnvironment object, config and index should both None.

        :param raw_config: Explicit configuration to use.  Either as a string (serialised in ini or yaml format) or
            a dictionary (deserialised).  If provided, config should be None.
            If an index or an explicit ODCEnvironment is supplied, config and raw_config should be None.

        :param app: A short, alphanumeric name to identify this application.

            The application name is used to track down problems with database queries, so it is strongly
            advised that be used.  Should be None if an index is supplied.

        :param validate_connection: Check that the database connection is available and valid.
            Defaults to True. Ignored if index is passed.
        """
        # Validate arguments

        if index is not None:
            # If an explicit index is provided, all other index-creation arguments should be None.
            should_be_none: list[str] = []
            if config is not None:
                should_be_none.append("config")
            if raw_config is not None:
                should_be_none.append("raw_config")
            if app is not None:
                should_be_none.append("app")
            if env is not None:
                should_be_none.append("env")
            if should_be_none:
                raise ValueError(
                    f"When an explicit index is provided, these arguments should be None: {','.join(should_be_none)}"
                )
            # Explicit index passed in?  Use it.
            self.index = index
            return

        # Obtain an ODCEnvironment object:
        cfg_env = ODCConfig.get_environment(
            env=env, config=config, raw_config=raw_config
        )

        self.index = index_connect(
            cfg_env, application_name=app, validate_connection=validate_connection
        )

    def list_products(
        self, with_pandas: bool = True, dataset_count: bool = False
    ) -> DataFrame | DataFrameLike:
        """
        List all products in the datacube. This will produce a ``pandas.DataFrame``
        or list of dicts containing useful information about each product, including:

            'name'
            'description'
            'license'
            'default_crs' or 'grid_spec.crs'
            'default_resolution' or 'grid_spec.crs'
            'dataset_count' (optional)

        :param with_pandas:
            Return the list as a Pandas DataFrame. Defaults to True.  If False, return a list of dicts.

        :param dataset_count:
            Return a "dataset_count" column containing the number of datasets
            for each product. This can take several minutes on large datacubes.
            Defaults to False.

        :return: A table or list of every product in the datacube.
        """

        def _get_non_default(product, col):
            load_hints = product.load_hints()
            if load_hints:
                if col == "crs":
                    return load_hints.get("output_crs", None)
                return load_hints.get(col, None)
            return getattr(product.grid_spec, col, None)

        # Read properties from each datacube product
        cols = [
            "name",
            "description",
            "license",
            "default_crs",
            "default_resolution",
        ]
        rows = [
            [
                (
                    getattr(pr, col, None)
                    # if 'default_crs' and 'default_resolution' are not None
                    # return 'default_crs' and 'default_resolution'
                    if getattr(pr, col, None) or "default" not in col
                    # else get crs and resolution from load_hints or grid_spec
                    # as per output_geobox() handling logic
                    else _get_non_default(pr, col.replace("default_", ""))
                )
                for col in cols
            ]
            for pr in self.index.products.get_all()
        ]

        # Optionally compute dataset count for each product and add to row/cols
        # Product lists are sorted by product name to ensure 1:1 match
        if dataset_count:
            # Load counts
            counts = [(p.name, c) for p, c in self.index.datasets.count_by_product()]

            # Sort both rows and counts by product name
            from operator import itemgetter

            rows = sorted(rows, key=itemgetter(0))
            counts = sorted(counts, key=itemgetter(0))

            # Add sorted count to each existing row
            rows = [row + [count[1]] for row, count in zip(rows, counts)]
            cols += ["dataset_count"]

        # If pandas not requested, return list of dicts
        if not with_pandas:
            return [dict(zip(cols, row)) for row in rows]

        # Return pandas dataframe with each product as a row
        import pandas

        return pandas.DataFrame(rows, columns=cols).set_index("name", drop=False)

    @deprecat.deprecat(
        deprecated_args={
            "show_archived": {
                "reason": "The show_archived argument has never done anything and will be removed in future.",
                "version": "1.9.0",
                "category": ODC2DeprecationWarning,
            }
        }
    )
    def list_measurements(
        self, show_archived: bool = False, with_pandas: bool = True
    ) -> DataFrame | DataFrameLike:
        """
        List measurements for each product

        :param show_archived: include archived products in the result.
        :param with_pandas: return the list as a Pandas DataFrame, otherwise as a list of dict. (defaults to True)
        """
        measurements = self._list_measurements()
        if not with_pandas:
            return measurements

        import pandas

        return pandas.DataFrame.from_records(measurements).set_index(
            ["product", "measurement"]
        )

    def _list_measurements(self) -> list[dict[str, Any]]:
        measurements = []
        dts = self.index.products.get_all()
        for dt in dts:
            if dt.measurements:
                for name, measurement in dt.measurements.items():
                    row = {
                        "product": dt.name,
                        "measurement": name,
                    }
                    if "attrs" in measurement:
                        row.update(measurement["attrs"])
                    row.update({k: v for k, v in measurement.items() if k != "attrs"})
                    measurements.append(row)
        return measurements

    #: pylint: disable=too-many-arguments, too-many-locals
    def load(
        self,
        product: str | None = None,
        measurements: str | list[str] | None = None,
        output_crs: MaybeCRS = None,
        resolution: (
            int
            | float
            | tuple[int | float, int | float]
            | list[int | float]
            | Resolution
            | None
        ) = None,
        resampling: Resampling | dict[str, Resampling] | None = None,
        align: XY[float] | Iterable[float] | None = None,
        skip_broken_datasets: bool | None = None,
        dask_chunks: Mapping[str, int | Literal["auto"]] | None = None,
        like: GeoBox | xarray.Dataset | xarray.DataArray | None = None,
        fuse_func: FuserFunction | Mapping[str, FuserFunction | None] | None = None,
        datasets: Sequence[Dataset] | None = None,
        dataset_predicate: Callable[[Dataset], bool] | None = None,
        progress_cbk: ProgressFunction | None = None,
        patch_url: Callable[[str], str] | None = None,
        limit: int | None = None,
        driver: Any | None = None,
        **query: QueryField,
    ) -> xarray.Dataset:
        r"""
        Load data as an ``xarray.Dataset`` object.
        Each measurement will be a data variable in the :class:`xarray.Dataset`.

        See the `xarray documentation <https://xarray.pydata.org/en/stable/data-structures.html>`_ for usage of the
        :class:`xarray.Dataset` and :class:`xarray.DataArray` objects.

        **Product and Measurements**

            A product can be specified using the product name::

                product='ls5_ndvi_albers'

            See :meth:`list_products` for the list of products with their names and properties.

            A product name MUST be supplied unless search is bypassed all together by supplying an explicit
            list of datasets.

            The ``measurements`` argument is a list of measurement names, as listed in :meth:`list_measurements`.
            If not provided, all measurements for the product will be returned::

                measurements=['red', 'nir', 'swir2']

        **Dimensions**

            Spatial dimensions can be specified using the ``longitude``/``latitude`` and ``x``/``y`` fields.

            The CRS of this query is assumed to be WGS84/EPSG:4326 unless the ``crs`` field is supplied,
            even if the stored data is in another projection or the ``output_crs`` is specified.
            The dimensions ``longitude``/``latitude`` and ``x``/``y`` can be used interchangeably::

                latitude=(-34.5, -35.2), longitude=(148.3, 148.7)

            or::

                x=(1516200, 1541300), y=(-3867375, -3867350), crs='EPSG:3577'

            You can also specify a polygon with an arbitrary CRS (in e.g. the native CRS)::

                geopolygon=polygon(coords, crs="EPSG:3577")

            Or an iterable of polygons (search is done against the union of all polygons)::

                geopolygon=[poly1, poly2, poly3, ....]

            You can also pass a WKT string, or a GeoJSON string or any other object that can be passed to the
            odc.geo.Geometry constructor, or an iterable of any of the above.

            Performance and accuracy of geopolygon queries may vary depending on the index driver in use and the CRS.

            The ``time`` dimension can be specified using a single or tuple of datetime objects or strings with
            ``YYYY-MM-DD hh:mm:ss`` format. Data will be loaded inclusive of the start and finish times.
            A ``None`` value in the range indicates an open range, with the provided date serving as either the
            upper or lower bound. E.g.::

                time=('2000-01-01', '2001-12-31')
                time=('2000-01', '2001-12')
                time=('2000', '2001')
                time=('2000')
                time=('2000', None)  # all data from 2000 onward
                time=(None, '2000')  # all data up to and including 2000

            For 3D datasets, where the product definition contains an ``extra_dimension`` specification,
            these dimensions can be queried using that dimension's name. E.g.::

                z=(10, 30)

            or::

                z=5

            or::

                wvl=(560.3, 820.5)

            For EO-specific datasets that are based around scenes, the time dimension can be reduced to the day level,
            using solar day to keep scenes together::

                group_by='solar_day'

            For data that has different values for the scene overlap that requires more complex rules for combining
            data, a function can be provided to the merging into a single time slice.

            See :func:`datacube.helpers.ga_pq_fuser` for an example implementation.
            see :func:`datacube.api.query.query_group_by` for `group_by` built-in functions.


        **Output**

            To reproject or resample data, supply the ``output_crs``, ``resolution``, ``resampling`` and ``align``
            fields.

            By default, the resampling method is 'nearest'. However, any stored overview layers may be used
            when down-sampling, which may override (or hybridise) the choice of resampling method.

            To reproject data to 30 m resolution for EPSG:3577::

                dc.load(product='ls5_nbar_albers',
                        x=(148.15, 148.2),
                        y=(-35.15, -35.2),
                        time=('1990', '1991'),
                        output_crs='EPSG:3577`,
                        resolution=30,
                        resampling='cubic'
                )

            odc-geo style xy objects are preferred for passing in resolution and align pairs to avoid x/y ordering
            ambiguity.

        :param product:
            The name of the product to be loaded. Either ``product`` or ``datasets`` must be supplied

        :param measurements:
            Measurements name or list of names to be included, as listed in :meth:`list_measurements`.
            These will be loaded as individual ``xr.DataArray`` variables in
            the output ``xarray.Dataset`` object.

            If a list is specified, the measurements will be returned in the order requested.
            By default, all available measurements are included.

        :param output_crs:
            The CRS of the returned data, for example ``EPSG:3577``.
            If no CRS is supplied, the CRS of the stored data is used if available.

            Any form that can be converted to a CRS by odc-geo is accepted.

            This differs from the ``crs`` parameter described above, which is used to define the CRS
            of the coordinates in the query itself.

        :param resolution:
            The spatial resolution of the returned data. If using square pixels with an inverted Y axis, it
            should be provided as an int or float. If not, it should be provided as an odc-geo XY object
            to avoid coordinate-order ambiguity.  If passed as a tuple, y,x order is assumed for backwards
            compatibility.

            Units are in the coordinate space of ``output_crs``. This includes the direction (as indicated by
            a positive or negative number).

        :param resampling:
            The resampling method to use if re-projection is required. This could be a string or
            a dictionary mapping band name to resampling mode. When using a dict use ``'\*'`` to
            indicate "apply to all other bands", for example ``{'\*': 'cubic', 'fmask': 'nearest'}`` would
            use ``cubic`` for all bands except ``fmask`` for which ``nearest`` will be used.

            Valid values are::

               'nearest', 'average', 'bilinear', 'cubic', 'cubic_spline',
               'lanczos', 'mode', 'gauss',  'max', 'min', 'med', 'q1', 'q3'

            Default is to use ``nearest`` for all bands.

            .. seealso::
               :meth:`load_data`

        :param align:
            Load data such that point 'align' lies on the pixel boundary.  A pair of floats between 0 and 1.

            An odc-geo XY object is preferred to avoid coordinate-order ambiguity.  If passed as a tuple, x,y
            order is assumed for backwards compatibility.

            Default is ``(0, 0)``

        :param skip_broken_datasets:
            Optional. If this is True, then don't break when failing to load a broken dataset.
            If None, the value will come from the environment variable of the same name.
            Default is False.

        :param dask_chunks:
            If the data should be lazily loaded using :class:`dask.array.Array`,
            specify the chunking size in each output dimension.

            See the documentation on using `xarray with dask <https://xarray.pydata.org/en/stable/dask.html>`_
            for more information.

        :param like:
            Use the output of a previous :meth:`load()` to load data into the same spatial grid and
            resolution (i.e. :class:`odc.geo.geobox.GeoBox` or an xarray `Dataset` or `DataArray`).
            E.g.::

                pq = dc.load(product='ls5_pq_albers', like=nbar_dataset)

        :param fuse_func: Function used to fuse/combine/reduce data with the ``group_by`` parameter. By default,
            data is simply copied over the top of each other in a relatively undefined manner. This function can
            perform a specific combining step. This can be a dictionary if different
            fusers are needed per band (similar format to the resampling dict described above).

        :param datasets: Optional. If this is a non-empty list of :class:`datacube.model.Dataset` objects,
            these will be loaded instead of performing a database lookup.

        :param dataset_predicate: Optional. A function that can be passed to restrict loaded datasets. A
            predicate function should take a :class:`datacube.model.Dataset` object (e.g. as returned from
            :meth:`find_datasets`) and return a boolean.

            For example, loaded data could be filtered to January observations only by passing the following
            predicate function that returns True for datasets acquired in January::

                def filter_jan(dataset): return dataset.time.begin.month == 1

            .

        :param progress_cbk: ``Int, Int -> None``,
            if supplied will be called for every file read with ``files_processed_so_far, total_files``. This is
            only applicable to non-lazy loads, ignored when using dask.

        :param patch_url: if supplied, will be used to patch/sign the url(s), as required to access some
            commercial archives (e.g. Microsoft Planetary Computer).

        :param limit: Optional. If provided, limit the maximum number of datasets returned. Useful for
            testing and debugging. Can also be provided via the ``dc_load_limit`` config option.

        :param driver: Optional. If provided, use the specified driver to load the data.

        :param query: Search parameters for products and dimension ranges as described above.
            For example: ``'x', 'y', 'time', 'crs'``.

        :return: Requested data in a :class:`xarray.Dataset`
        """
        if product is None and datasets is None:
            raise ValueError("Must specify a product or supply datasets")

        if datasets is None:
            assert product is not None  # For type checker
            if limit is None:
                # check if a value was provided via the envvar
                limit = self.index.environment["dc_load_limit"]
            datasets = self.find_datasets(
                ensure_location=True,
                dataset_predicate=dataset_predicate,
                like=like,
                limit=limit,
                product=product,
                **query,
            )
        elif isinstance(datasets, collections.abc.Iterator):
            datasets = list(datasets)

        if len(datasets) == 0:
            return xarray.Dataset()

        ds, *_ = datasets
        datacube_product = ds.product

        # Retrieve extra_dimension from product definition
        extra_dims: ExtraDimensions | None = None
        if datacube_product:
            extra_dims = datacube_product.extra_dimensions

            # Extract extra_dims slice information
            extra_dims_slice: ExtraDimensionSlices = {
                k: v  # type: ignore[misc]
                for k, v in query.items()
                if v is not None and k in extra_dims.dims
            }
            extra_dims = extra_dims[extra_dims_slice]
            # Check if empty
            if extra_dims.has_empty_dim():
                return xarray.Dataset()

        if isinstance(resolution, tuple | list):
            resolution = _handle_legacy_resolution(resolution)

        load_hints = datacube_product.load_hints()
        grid_spec = None if load_hints is not None else datacube_product.grid_spec

        geobox = output_geobox(
            like=like,
            output_crs=output_crs,
            resolution=resolution,
            align=align,
            grid_spec=grid_spec,
            load_hints=load_hints,
            datasets=datasets,
            geopolygon=cast(Geometry | None, query.pop("geopolygon", None)),
            **query,
        )
        group_by = query_group_by(**query)  # type: ignore[arg-type]
        grouped = self.group_datasets(datasets, group_by)

        measurement_dicts = datacube_product.lookup_measurements(measurements)

        if skip_broken_datasets is None:
            # default to value from env var, which defaults to False
            skip_broken_datasets = self.index.environment["skip_broken_datasets"]

        result = self.load_data(
            grouped,
            geobox,
            measurement_dicts,
            resampling=resampling,
            fuse_func=fuse_func,
            dask_chunks=dask_chunks,
            skip_broken_datasets=skip_broken_datasets,
            progress_cbk=progress_cbk,
            extra_dims=extra_dims,
            patch_url=patch_url,
            driver=driver,
        )

        return result

    def find_datasets(
        self,
        ensure_location: bool = False,
        dataset_predicate: Callable[[Dataset], bool] | None = None,
        like: GeoBox | xarray.Dataset | xarray.DataArray | None = None,
        limit: int | None = None,
        **search_terms: QueryField,
    ) -> list[Dataset]:
        """
        Search the index and return all datasets for a product matching the search terms.

        :param ensure_location: only return datasets that have locations
        :param dataset_predicate: an optional predicate to filter datasets
        :param like:
            Use the output of a previous :meth:`load()` to load data into the same spatial grid and
            resolution (i.e. :class:`odc.geo.geobox.GeoBox` or an xarray `Dataset` or `DataArray`).
            E.g.::

                pq = dc.load(product='ls5_pq_albers', like=nbar_dataset)

        :param limit: if provided, limit the maximum number of datasets returned
        :param search_terms: see :class:`datacube.api.query.Query`
        :return: list of datasets

        .. seealso:: :meth:`group_datasets` :meth:`load_data` :meth:`find_datasets_lazy`
        """
        return list(
            self.find_datasets_lazy(
                limit=limit,
                ensure_location=ensure_location,
                dataset_predicate=dataset_predicate,
                like=like,
                **search_terms,
            )
        )

    def find_datasets_lazy(
        self,
        limit: int | None = None,
        ensure_location: bool = False,
        dataset_predicate: Callable[[Dataset], bool] | None = None,
        like: GeoBox | xarray.Dataset | xarray.DataArray | None = None,
        **kwargs: QueryField,
    ) -> Iterable[Dataset]:
        """
        Find datasets matching query.

        :param limit: if provided, limit the maximum number of datasets returned
        :param ensure_location: only return datasets that have locations
        :param dataset_predicate: an optional predicate to filter datasets
        :param like:
            Use the output of a previous :meth:`load()` to load data into the same spatial grid and
            resolution (i.e. :class:`odc.geo.geobox.GeoBox` or an xarray `Dataset` or `DataArray`).
            E.g.::

                pq = dc.load(product='ls5_pq_albers', like=nbar_dataset)
        :param kwargs: see :class:`datacube.api.query.Query`
        :return: iterator of datasets

        .. seealso:: :meth:`group_datasets` :meth:`load_data` :meth:`find_datasets`
        """
        if like is not None:
            like = _normalise_geobox(like)

        query = Query(self.index, like=like, **kwargs)  # type: ignore[arg-type]
        if not query.product:
            raise ValueError("must specify a product")

        datasets = self.index.datasets.search(limit=limit, **query.search_terms)

        if query.geopolygon is not None and not self.index.supports_spatial_indexes:
            datasets = select_datasets_inside_polygon(datasets, query.geopolygon)

        if ensure_location:
            datasets = (dataset for dataset in datasets if dataset.uri)

        # If a predicate function is provided, use this to filter datasets before load
        if dataset_predicate is not None:
            datasets = (dataset for dataset in datasets if dataset_predicate(dataset))

        return datasets

    @staticmethod
    def group_datasets(
        datasets: Iterable[Dataset], group_by: GroupBy
    ) -> xarray.DataArray:
        """
        Group datasets along defined non-spatial dimensions (ie. time).

        :param datasets: a list of datasets, typically from :meth:`find_datasets`
        :param group_by: Contains:
            - a function that returns a label for a dataset
            - name of the new dimension
            - unit for the new dimension
            - function to sort by before grouping

        .. seealso:: :meth:`find_datasets`, :meth:`load_data`, :meth:`query_group_by`
        """
        if isinstance(group_by, str):
            group_by = query_group_by(group_by=group_by)

        def ds_sorter(ds: Dataset) -> Any:
            return group_by.sort_key(ds), getattr(ds, "id", 0)

        def norm_axis_value(x: Any) -> Any:
            if isinstance(x, datetime.datetime):
                # For datetime we convert to UTC, then strip timezone info
                # to avoid numpy/pandas warning about timezones
                return numpy.datetime64(normalise_dt(x), "ns")
            return x

        def mk_group(group: Iterable[Dataset]) -> tuple[Any, Iterable[Dataset]]:
            dss = tuple(sorted(group, key=ds_sorter))
            return norm_axis_value(group_by.group_key(dss)), dss

        datasets = sorted(datasets, key=group_by.group_by_func)

        groups = [
            mk_group(group) for _, group in groupby(datasets, group_by.group_by_func)
        ]

        groups.sort(key=lambda x: x[0])

        coords = numpy.asarray([coord for coord, _ in groups])
        data = numpy.empty(len(coords), dtype=object)
        for i, (_, dss) in enumerate(groups):
            data[i] = dss

        sources = xarray.DataArray(data, dims=[group_by.dimension], coords=[coords])
        if coords.dtype.kind == "M":
            # skip units for time dimensions as it breaks .to_netcdf(..) functionality #972
            sources[group_by.dimension].attrs["units"] = group_by.units

        return sources

    @staticmethod
    def create_storage(
        coords: DataArrayCoordinates | None,
        geobox: GeoBox | xarray.Dataset | xarray.DataArray,
        measurements: list[Measurement],
        data_func: (
            Callable[[Measurement, tuple[int, ...]], numpy.ndarray] | None
        ) = None,
        extra_dims: ExtraDimensions | None = None,
    ) -> xarray.Dataset:
        """
        Create a :class:`xarray.Dataset` and (optionally) fill it with data.

        This function makes the in memory storage structure to hold datacube data.

        :param coords:
            DataArrayCoordinates defining the dimensions not specified by `geobox`

        :param geobox:
            A GeoBox defining the output spatial projection and resolution

        :param measurements:
            list of :class:`datacube.model.Measurement`

        :param data_func: Callable `Measurement -> np.ndarray`
            function to fill the storage with data. It is called once for each measurement, with the measurement
            as an argument. It should return an appropriately shaped numpy array. If not provided memory is
            allocated and filled with `nodata` value defined on a given Measurement.

        :param extra_dims:
            A ExtraDimensions describing any additional dimensions on top of (t, y, x)


        .. seealso:: :meth:`find_datasets` :meth:`group_datasets`
        """
        from collections import OrderedDict
        from copy import deepcopy

        spatial_ref = "spatial_ref"

        def empty_func(m: Measurement, shape: tuple[int, ...]) -> numpy.ndarray:
            return numpy.full(shape, m.nodata, dtype=m.dtype)

        geobox = _normalise_geobox(geobox)

        crs_attrs = {}
        if geobox.crs is not None:
            crs_attrs["crs"] = str(geobox.crs)
            crs_attrs["grid_mapping"] = spatial_ref

        # Assumptions
        #  - 3D dims must fit between (t) and (y, x) or (lat, lon)

        # 2D defaults
        # retrieve dims from coords if DataArray
        dims_default: tuple[Hashable, ...] = ()
        if coords is not None:
            coords_value = next(iter(coords.values()))
            if isinstance(coords_value, xarray.DataArray):
                dims_default = coords_value.dims + geobox.dimensions

        if not dims_default:
            dims_default = (() if coords is None else coords.dims) + geobox.dimensions

        shape_default = (
            ()
            if coords is None
            else tuple(c.size for k, c in coords.items() if k in dims_default)
        ) + geobox.shape
        coords_default: OrderedDict[str, xarray.DataArray] = OrderedDict()
        if coords is not None:
            coords_default.update([(str(k), v) for k, v in coords.items()])
        coords_default.update(
            [(str(k), v) for k, v in xr_coords(geobox, spatial_ref).items()]
        )

        arrays = []
        ds_coords = deepcopy(coords_default)

        for m in measurements:
            if "extra_dim" not in m:
                # 2D default case
                arrays.append((m, shape_default, coords_default, dims_default))
            elif extra_dims:
                # 3D case
                name = m.extra_dim
                new_dims = dims_default[:1] + (name,) + dims_default[1:]
                new_coords = deepcopy(coords_default)
                new_coords[name] = extra_dims._coords[name].copy()
                new_coords[name].attrs.update(crs_attrs)
                ds_coords.update(new_coords)

                new_shape = (
                    shape_default[:1]
                    + (len(new_coords[name].values),)
                    + shape_default[1:]
                )
                arrays.append((m, new_shape, new_coords, new_dims))

        data_func = data_func or (lambda m, shape: empty_func(m, shape))

        def mk_data_var(
            m: Measurement,
            shape: tuple[int, ...],
            coords: OrderedDict[str, xarray.DataArray],
            dims: tuple[Hashable, ...],
            data_func: Callable[[Measurement, tuple[int, ...]], numpy.ndarray],
        ) -> xarray.DataArray:
            data = data_func(m, shape)
            attrs = dict(**m.dataarray_attrs(), **crs_attrs)
            return xarray.DataArray(
                data, name=m.name, coords=coords, dims=dims, attrs=attrs
            )

        return xarray.Dataset(
            {
                m.name: mk_data_var(m, shape, coords, dims, data_func)
                for m, shape, coords, dims in arrays
            },
            coords=ds_coords,
            attrs=crs_attrs,
        )

    @staticmethod
    def _dask_load(
        sources: xarray.DataArray,
        geobox: GeoBox,
        measurements: list[Measurement],
        dask_chunks: Mapping[str, int | Literal["auto"]],
        skip_broken_datasets: bool = False,
        extra_dims: ExtraDimensions | None = None,
        patch_url: Callable[[str], str] | None = None,
    ) -> xarray.Dataset:
        chunk_sizes = _calculate_chunk_sizes(sources, geobox, dask_chunks, extra_dims)
        needed_irr_chunks = chunk_sizes[0]
        if extra_dims:
            extra_dim_chunks = chunk_sizes[1]
        grid_chunks = chunk_sizes[-1]
        gbt = GeoboxTiles(geobox, grid_chunks)
        dsk = {}

        def chunk_datasets(dss, gbt):
            out = {}
            for ds in dss:
                dsk[_tokenize_dataset(ds)] = ds
                for idx in gbt.tiles(ds.extent):
                    out.setdefault(idx, []).append(ds)
            return out

        chunked_srcs = xr_apply(
            sources, lambda _, dss: chunk_datasets(dss, gbt), dtype=object
        )

        def data_func(measurement, shape):
            if "extra_dim" in measurement:
                chunks = needed_irr_chunks + extra_dim_chunks + grid_chunks
            else:
                chunks = needed_irr_chunks + grid_chunks
            return _make_dask_array(
                chunked_srcs,
                dsk,
                gbt,
                measurement,
                chunks=chunks,
                skip_broken_datasets=skip_broken_datasets,
                extra_dims=extra_dims,
                patch_url=patch_url,
            )

        return Datacube.create_storage(
            sources.coords, geobox, measurements, data_func, extra_dims
        )

    @staticmethod
    def _xr_load(
        sources: xarray.DataArray,
        geobox: GeoBox,
        measurements: list[Measurement],
        skip_broken_datasets: bool = False,
        progress_cbk: ProgressFunction | None = None,
        extra_dims: ExtraDimensions | None = None,
        patch_url: Callable[[str], str] | None = None,
    ) -> xarray.Dataset:
        def mk_cbk(cbk: ProgressFunction | None) -> ProgressFunction | None:
            if cbk is None:
                return None
            n = 0
            t_size = sum(len(x) for x in sources.values.ravel())
            n_total = 0
            for m in measurements:
                if "extra_dim" in m:
                    assert extra_dims is not None  # for type-checker
                    index_subset = extra_dims.measurements_slice(m.extra_dim)
                    n_total += t_size * len(
                        m.extra_dim.get("measurement_map")[index_subset]
                    )
                else:
                    n_total += t_size

            def _cbk(_a: int, _b: int) -> Any | None:
                nonlocal n
                n += 1
                return cbk(n, n_total)

            return _cbk

        data = Datacube.create_storage(
            sources.coords, geobox, measurements, extra_dims=extra_dims
        )
        _cbk = mk_cbk(progress_cbk)

        # Create a list of read IO operations
        read_ios = []
        for index, datasets in numpy.ndenumerate(sources.values):
            for m in measurements:
                if "extra_dim" in m:
                    # When we want to support 3D native reads, we can start by replacing the for loop with
                    # read_ios.append(((index + extra_dim_index), (datasets, m, index_subset)))
                    assert extra_dims is not None  # for type-checker
                    index_subset = extra_dims.measurements_index(m.extra_dim)
                    for result_index, extra_dim_index in enumerate(
                        range(*index_subset)
                    ):
                        read_ios.append(
                            ((index + (result_index,)), (datasets, m, extra_dim_index))
                        )
                else:
                    # Get extra_dim index if available
                    extra_dim_index = m.get("extra_dim_index", None)
                    read_ios.append((index, (datasets, m, extra_dim_index)))

        # Perform the read IO operations
        for index, (datasets, m, extra_dim_index) in read_ios:
            data_slice = data[m.name].values[index]
            try:
                _fuse_measurement(
                    data_slice,
                    datasets,
                    geobox,
                    m,
                    skip_broken_datasets=skip_broken_datasets,
                    progress_cbk=_cbk,
                    extra_dim_index=extra_dim_index,
                    patch_url=patch_url,
                )
            except (TerminateCurrentLoad, KeyboardInterrupt):
                data.attrs["dc_partial_load"] = True
                return data

        return data

    @staticmethod
    def load_data(
        sources: xarray.DataArray,
        geobox: GeoBox | xarray.Dataset | xarray.DataArray,
        measurements: Mapping[str, Measurement] | list[Measurement],
        resampling: Resampling | dict[str, Resampling] | None = None,
        fuse_func: FuserFunction | Mapping[str, FuserFunction | None] | None = None,
        dask_chunks: Mapping[str, int | Literal["auto"]] | None = None,
        skip_broken_datasets: bool = False,
        progress_cbk: ProgressFunction | None = None,
        extra_dims: ExtraDimensions | None = None,
        patch_url: Callable[[str], str] | None = None,
        driver: Any | None = None,
        **extra,
    ) -> xarray.Dataset:
        """
        Load data from :meth:`group_datasets` into an :class:`xarray.Dataset`.

        :param sources:
            DataArray holding a list of :class:`datacube.model.Dataset`, grouped along the time dimension

        :param geobox:
            A GeoBox defining the output spatial projection and resolution

        :param measurements:
            list of `Measurement` objects

        :param resampling:
            The resampling method to use if re-projection is required. This could be a string or
            a dictionary mapping band name to resampling mode. When using a dict use ``'*'`` to
            indicate "apply to all other bands", for example ``{'*': 'cubic', 'fmask': 'nearest'}`` would
            use `cubic` for all bands except ``fmask`` for which `nearest` will be used.

            Valid values are: ``'nearest', 'cubic', 'bilinear', 'cubic_spline', 'lanczos', 'average',
            'mode', 'gauss',  'max', 'min', 'med', 'q1', 'q3'``

            Default is to use ``nearest`` for all bands.

        :param fuse_func:
            function to merge successive arrays as an output. Can be a dictionary just like resampling.

        :param dask_chunks:
            If provided, the data will be loaded on demand using :class:`dask.array.Array`.
            Should be a dictionary specifying the chunking size for each output dimension.
            Unspecified dimensions will be auto-guessed, currently this means use chunk size of 1 for non-spatial
            dimensions and use whole dimension (no chunking unless specified) for spatial dimensions.

            See the documentation on using `xarray with dask <https://xarray.pydata.org/en/stable/dask.html>`_
            for more information.

        :param skip_broken_datasets: do not include broken datasets in the result.

        :param progress_cbk: Int, Int -> None
            if supplied will be called for every file read with `files_processed_so_far, total_files`. This is
            only applicable to non-lazy loads, ignored when using dask.

        :param extra_dims:
            A ExtraDimensions describing any additional dimensions on top of (t, y, x)

        :param patch_url:
            if supplied, will be used to patch/sign the url(s), as required to access some commercial archives.

        :param driver:
            Optional. If provided, use the specified driver to load the data.


        .. seealso:: :meth:`find_datasets` :meth:`group_datasets`
        """
        measurements = per_band_load_data_settings(
            measurements, resampling=resampling, fuse_func=fuse_func
        )

        geobox = _normalise_geobox(geobox)

        if driver is not None:
            from ..storage._loader import driver_based_load

            return driver_based_load(
                driver,
                sources,
                geobox,
                measurements,
                dask_chunks,
                skip_broken_datasets=skip_broken_datasets,
                extra_dims=extra_dims,
                patch_url=patch_url,
            )

        if dask_chunks is not None:
            return Datacube._dask_load(
                sources,
                geobox,
                measurements,
                dask_chunks,
                skip_broken_datasets=skip_broken_datasets,
                extra_dims=extra_dims,
                patch_url=patch_url,
            )
        else:
            return Datacube._xr_load(
                sources,
                geobox,
                measurements,
                skip_broken_datasets=skip_broken_datasets,
                progress_cbk=progress_cbk,
                extra_dims=extra_dims,
                patch_url=patch_url,
            )

    @override
    def __str__(self) -> str:
        return f"Datacube<index={self.index!r}>"

    @override
    def __repr__(self) -> str:
        return self.__str__()

    def close(self) -> None:
        """
        Close any open connections
        """
        self.index.close()

    def __enter__(self) -> Datacube:
        return self

    def __exit__(self, type_, value, traceback) -> None:
        self.close()


def per_band_load_data_settings(
    measurements: list[Measurement] | Mapping[str, Measurement],
    resampling: Resampling | Mapping[str, Resampling] | None = None,
    fuse_func: FuserFunction | Mapping[str, FuserFunction | None] | None = None,
) -> list[Measurement]:
    def with_resampling(m, resampling, default=None):
        m = m.copy()
        m["resampling_method"] = resampling.get(m.name, default)
        return m

    def with_fuser(m, fuser, default=None):
        m = m.copy()
        m["fuser"] = fuser.get(m.name, default)
        return m

    if resampling is not None and not isinstance(resampling, dict):
        resampling = {"*": resampling}

    if fuse_func is None or callable(fuse_func):
        fuse_func = {"*": fuse_func}

    if isinstance(measurements, dict):
        measurements = list(measurements.values())

    if resampling is not None:
        measurements = [
            with_resampling(m, resampling, default=resampling.get("*"))
            for m in measurements
        ]

    if fuse_func is not None:
        measurements = [
            with_fuser(m, fuse_func, default=fuse_func.get("*")) for m in measurements
        ]

    return measurements


def output_geobox(
    like: GeoBox | LegacyGeoBox | xarray.Dataset | xarray.DataArray | None = None,
    output_crs: Any = None,
    resolution: (
        int
        | float
        | Resolution
        | tuple[int | float, int | float]
        | list[int | float]
        | None
    ) = None,
    align: XY[float] | Iterable[float] | None = None,
    grid_spec: GridSpec | None = None,
    load_hints: Mapping[str, Any] | None = None,
    datasets: Iterable[Dataset] | None = None,
    geopolygon: Geometry | None = None,
    **query: QueryField,
) -> GeoBox:
    """Configure output geobox from user provided output specs."""

    if like is not None:
        assert output_crs is None, "'like' and 'output_crs' are not supported together"
        assert resolution is None, "'like' and 'resolution' are not supported together"
        assert align is None, "'like' and 'align' are not supported together"
        return _normalise_geobox(like)

    if load_hints:
        if output_crs is None:
            output_crs = load_hints.get("output_crs", None)

        if resolution is None:
            resolution = load_hints.get("resolution", None)

        if align is None:
            align = load_hints.get("align", None)

    if output_crs is not None:
        if resolution is None:
            raise ValueError("Must specify 'resolution' when specifying 'output_crs'")
        crs = CRS(output_crs)
    elif grid_spec is not None:
        # specification from grid_spec
        crs = grid_spec.crs
        if resolution is None:
            resolution = grid_spec.resolution
        align = align or grid_spec.alignment
    else:
        raise ValueError(
            "Product has no default CRS.\nMust specify 'output_crs' and 'resolution'"
        )

    # Try figuring out bounds
    #  1. Explicitly defined with geopolygon
    #  2. Extracted from x=,y=
    #  3. Computed from dataset footprints
    #  4. fail with ValueError
    if geopolygon is None:
        geopolygon = extract_geom_from_query(**query)

        if geopolygon is None:
            if datasets is None:
                raise ValueError("Bounds are not specified")

            geopolygon = get_bounds(datasets, crs)

    if isinstance(resolution, tuple | list):
        resolution = _handle_legacy_resolution(resolution)

    if align is not None:
        align = yx_(align)

    return GeoBox.from_geopolygon(geopolygon, resolution, crs, align)


def select_datasets_inside_polygon(
    datasets: Iterable[Dataset], polygon: Geometry
) -> Iterable[Dataset]:
    # Check against the bounding box of the original scene, can throw away some portions
    # (Only needed for index drivers without spatial index support)
    query_crs = polygon.crs
    for dataset in datasets:
        if dataset.extent and intersects(polygon, dataset.extent.to_crs(query_crs)):
            yield dataset


def fuse_lazy(
    datasets: Iterable[Dataset],
    geobox: GeoBox,
    measurement: Measurement,
    skip_broken_datasets: bool = False,
    prepend_dims: int = 0,
    extra_dim_index: int | None = None,
    patch_url: Callable[[str], str] | None = None,
) -> numpy.ndarray:
    prepend_shape = (1,) * prepend_dims
    data = numpy.full(geobox.shape, measurement.nodata, dtype=measurement.dtype)
    _fuse_measurement(
        data,
        datasets,
        geobox,
        measurement,
        skip_broken_datasets=skip_broken_datasets,
        extra_dim_index=extra_dim_index,
        patch_url=patch_url,
    )
    return data.reshape(prepend_shape + geobox.shape)


def _fuse_measurement(
    dest: numpy.ndarray,
    datasets: Iterable[Dataset],
    geobox: GeoBox,
    measurement: Measurement,
    skip_broken_datasets: bool = False,
    progress_cbk: ProgressFunction | None = None,
    extra_dim_index: int | None = None,
    patch_url: Callable[[str], str] | None = None,
) -> None:
    srcs = []
    for ds in datasets:
        src = None
        with ignore_exceptions_if(skip_broken_datasets):
            src = new_datasource(BandInfo(ds, measurement.name, patch_url=patch_url))

        if src is None:
            if not skip_broken_datasets:
                raise ValueError(f"Failed to load dataset: {ds.id}")
        else:
            srcs.append(src)

    reproject_and_fuse(
        srcs,
        dest,
        geobox,
        dest.dtype.type(measurement.nodata),
        resampling=measurement.get("resampling_method", "nearest"),
        fuse_func=measurement.get("fuser", None),
        skip_broken_datasets=skip_broken_datasets,
        progress_cbk=progress_cbk,
        extra_dim_index=extra_dim_index,
    )


def get_bounds(datasets: Iterable[Dataset], crs: CRS) -> Geometry:
    bbox = bbox_union(ds.extent.to_crs(crs).boundingbox for ds in datasets if ds.extent)
    return box(*bbox, crs=crs)  # type: ignore[misc]


def _calculate_chunk_sizes(
    sources: xarray.DataArray,
    geobox: GeoBox,
    dask_chunks: Mapping[str, int | Literal["auto"]],
    extra_dims: ExtraDimensions | None = None,
) -> tuple[tuple, ...]:
    extra_dim_names: tuple[str, ...] = ()
    extra_dim_shapes: tuple[int, ...] = ()
    if extra_dims is not None:
        extra_dim_names, extra_dim_shapes = extra_dims.chunk_size()

    valid_keys = (
        tuple(str(dim) for dim in sources.dims) + extra_dim_names + geobox.dimensions
    )
    bad_keys = dask_chunks.keys() - set(valid_keys)
    if bad_keys:
        raise KeyError(
            f"Unknown dask_chunk dimension {bad_keys}. Valid dimensions are: {valid_keys}"
        )

    chunk_maxsz = dict(
        zip(
            sources.dims + extra_dim_names + geobox.dimensions,
            sources.shape + extra_dim_shapes + geobox.shape,
        )
    )

    # defaults: 1 for non-spatial, whole dimension for Y/X
    chunk_defaults = dict(
        [(dim, 1) for dim in sources.dims]
        + [(dim, 1) for dim in extra_dim_names]
        + [(dim, -1) for dim in geobox.dimensions]
    )

    def _resolve(k, v: str | int | None) -> int:
        if v is None or v == "auto":
            v = _resolve(k, chunk_defaults[k])

        if isinstance(v, int):
            if v < 0:
                return chunk_maxsz[k]
            return v
        raise ValueError("Chunk should be one of int|'auto'")

    irr_chunks = tuple(_resolve(dim, dask_chunks.get(str(dim))) for dim in sources.dims)
    extra_dim_chunks = tuple(
        _resolve(dim, dask_chunks.get(str(dim))) for dim in extra_dim_names
    )
    grid_chunks = tuple(
        _resolve(dim, dask_chunks.get(str(dim))) for dim in geobox.dimensions
    )

    if extra_dim_chunks:
        return irr_chunks, extra_dim_chunks, grid_chunks
    else:
        return irr_chunks, grid_chunks


def _tokenize_dataset(dataset: Dataset) -> str:
    return f"dataset-{dataset.id.hex}"


# pylint: disable=too-many-locals
def _make_dask_array(
    chunked_srcs: xarray.DataArray,
    dsk,
    gbt,
    measurement: Measurement,
    chunks,
    skip_broken_datasets: bool = False,
    extra_dims: ExtraDimensions | None = None,
    patch_url: Callable[[str], str] | None = None,
):
    dsk = dsk.copy()  # this contains mapping from dataset id to dataset object

    token = uuid.uuid4().hex
    dsk_name = f"dc_load_{measurement.name}-{token}"

    needed_irr_chunks, grid_chunks = chunks[:-2], chunks[-2:]
    actual_irr_chunks = (1,) * len(needed_irr_chunks)

    # we can have up to 4 empty chunk shapes: whole, right edge, bottom edge and
    # bottom right corner
    #  W R
    #  B BR
    empties: dict[tuple[int, int], str] = {}

    def _mk_empty(shape: tuple[int, int]) -> str:
        name = empties.get(shape)
        if name is not None:
            return name

        name = "empty_{}x{}-{token}".format(*shape, token=token)
        dsk[name] = (
            numpy.full,
            actual_irr_chunks + shape,
            measurement.nodata,
            measurement.dtype,
        )
        empties[shape] = name

        return name

    for irr_index, tiled_dss in numpy.ndenumerate(chunked_srcs.values):
        key_prefix = (dsk_name, *irr_index)

        # all spatial chunks
        for idx in numpy.ndindex(gbt.shape.shape):
            dss = tiled_dss.get(idx, None)

            if dss is None:
                val3d = _mk_empty(gbt.chunk_shape(idx).yx)
                # 3D case
                if "extra_dim" in measurement:
                    assert extra_dims is not None  # For type checker
                    index_subset = extra_dims.measurements_index(measurement.extra_dim)
                    for result_index, _ in numpy.ndenumerate(range(*index_subset)):
                        dsk[key_prefix + result_index + idx] = val3d
                else:
                    dsk[key_prefix + idx] = val3d
            else:
                val = (
                    fuse_lazy,
                    [_tokenize_dataset(ds) for ds in dss],
                    gbt[idx],
                    measurement,
                    skip_broken_datasets,
                    len(needed_irr_chunks),
                )

                # 3D case
                if "extra_dim" in measurement:
                    # Do extra_dim subsetting here
                    assert extra_dims is not None  # For type checker
                    index_subset = extra_dims.measurements_index(measurement.extra_dim)
                    for result_index, extra_dim_index in enumerate(
                        range(*index_subset)
                    ):
                        dsk[key_prefix + (result_index,) + idx] = val + (
                            extra_dim_index,
                            patch_url,
                        )
                else:
                    # Get extra_dim index if available
                    extra_dim_index = measurement.get("extra_dim_index", None)
                    dsk[key_prefix + idx] = val + (extra_dim_index, patch_url)

    y_shapes = [grid_chunks[0]] * gbt.shape[0]
    x_shapes = [grid_chunks[1]] * gbt.shape[1]

    y_shapes[-1], x_shapes[-1] = gbt.chunk_shape(tuple(n - 1 for n in gbt.shape))

    extra_dim_shape: tuple = ()
    if "extra_dim" in measurement:
        assert extra_dims is not None  # For type checker
        dim_name = measurement.extra_dim
        extra_dim_shape += (len(extra_dims.measurements_values(dim_name)),)

    data = da.Array(
        dsk,
        dsk_name,
        chunks=actual_irr_chunks + (tuple(y_shapes), tuple(x_shapes)),
        dtype=measurement.dtype,
        shape=(chunked_srcs.shape + extra_dim_shape + gbt.base.shape),
    )

    if needed_irr_chunks != actual_irr_chunks:
        data = data.rechunk(chunks=chunks)
    return data


def _handle_legacy_resolution(
    resolution: tuple[int | float, int | float] | list[int | float],
) -> int | float | Resolution | None:
    warnings.warn(
        "The use of tuples or lists for resolution is deprecated. "
        "Square resolutions can be provided as an int or float, "
        "or axis order can be specified with odc.geo.resxy_ or odc.geo.resyx_. "
        "Legacy resolution formats are assumed to use (y, x) ordering.",
        ODC2DeprecationWarning,
    )
    if not len(resolution):
        _LOG.warning("Empty resolution value. Ignoring")
        return None
    if len(resolution) == 1:
        return resolution[0]
    if len(resolution) > 2:
        raise ValueError("Resolution cannot have more than 2 dimensions.")
    if resolution[0] == -resolution[1]:
        return res_(resolution[1])
    return resyx_(*resolution)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Low level config path resolution, loading and multi-format parsing functions.

The default search path and default config also live here.
"""

from __future__ import annotations

import os
import warnings
from enum import Enum
from os import PathLike
from os.path import expanduser
from typing import TYPE_CHECKING

from datacube.cfg.exceptions import ConfigException
from datacube.cfg.utils import ConfigDict, SemaphoreCallback, smells_like_ini
from datacube.migration import ODC2DeprecationWarning

if TYPE_CHECKING:
    from datacube.cfg.api import GeneralisedPath

_DEFAULT_CONFIG_SEARCH_PATH: list[str] = [
    "datacube.conf",  # i.e. in the current working directory.
    expanduser("~/.datacube.conf"),  # i.e. in user's home directory.
    # Check if we are running under a Windows and use Windowsy default paths?
    "/etc/default/datacube.conf",  # Preferred location for global config file
    "/etc/datacube.conf",  # Legacy location for global config file
]

_DEFAULT_CONF = """
default:
   db_hostname: ''
   db_database: datacube
   index_driver: default
   db_connection_timeout: 60
"""


def find_config(
    paths_in: GeneralisedPath | None,
    default_cb: SemaphoreCallback | None = None,
) -> str:
    """
    Given a file system path, or a list of file system paths, return the contents of the first file
    in the list that can be read as a string.

    If "None" is passed in the default config search path is used:

        "datacube.conf"               (i.e. 'datacube.conf' in the current working directory.)
        "~/.datacube.conf"            (i.e. '.datacube.conf' in the user's home directory.)
        "/etc/default/datacube.conf"  (Preferred location for global config file)
        "/etc/datacube.conf"          (Legacy location for global config file)

    If a path or list of paths was passed in, AND no readable file could be found, a ConfigException is raised.
    If None was passed in, AND no readable file could be found, a default configuration text is returned.

    :param paths_in: A file system path, or a list of file system paths, or None.
    :param default_cb: A default semaphore callback object, or None.
    :return: The contents of the first readable file found.
    """
    using_default_paths: bool = False
    paths: list[str | PathLike] = []
    if paths_in is None:
        if os.environ.get("ODC_CONFIG_PATH"):
            paths.extend(os.environ["ODC_CONFIG_PATH"].split(":"))
        elif os.environ.get("DATACUBE_CONFIG_PATH"):
            warnings.warn(
                "Datacube config path being determined by legacy $DATACUBE_CONFIG_PATH environment variable. "
                "This environment variable is deprecated and the behaviour of it has changed somewhat since datacube "
                "1.8.x.   Please refer to the documentation for details and switch to $ODC_CONFIG_PATH",
                ODC2DeprecationWarning,
            )
            paths.extend(os.environ["DATACUBE_CONFIG_PATH"].split(":"))
        else:
            paths.extend(_DEFAULT_CONFIG_SEARCH_PATH)
            using_default_paths = True
    elif isinstance(paths_in, str | PathLike):
        paths.append(paths_in)
    else:
        paths.extend(paths_in)

    for path in paths:
        try:
            with open(path) as fp:
                return fp.read()
        except OSError:
            continue

    if using_default_paths:
        if default_cb is not None:
            default_cb()
        return _DEFAULT_CONF

    raise ConfigException("No configuration file found in the provided locations")


class CfgFormat(Enum):
    """
    An Enum class for config file formats.
    """

    AUTO = 0  # Format unspecified - autodetect
    INI = 1
    YAML = 2
    JSON = 2  # JSON is a subset of YAML


def parse_text(cfg_text: str, fmt: CfgFormat = CfgFormat.AUTO) -> ConfigDict:
    """
    Parse a string of text in INI, JSON or YAML format into a raw dictionary.

    Raises a ConfigException if the file cannot be parsed.

    :param cfg_text: Configuration string in INI, YAML or JSON format
    :param fmt: Whether to use the ini or yaml/json parser. Autodetects file format by default.
    :return: A raw config dictionary
    """
    raw_config = {}
    if fmt == fmt.INI or (fmt == fmt.AUTO and smells_like_ini(cfg_text)):
        # INI parsing
        import configparser

        try:
            ini_config = configparser.ConfigParser()
            ini_config.read_string(cfg_text)
            for section in ini_config.sections():
                sect = {}
                for key, value in ini_config.items(section):
                    sect[key] = value

                raw_config[section] = sect
        except configparser.Error as e:
            raise ConfigException(f"Invalid INI file: {e}") from None
    else:
        # YAML/JSON parsing
        import yaml

        try:
            raw_config = yaml.load(cfg_text, Loader=yaml.Loader)
        except yaml.parser.ParserError as e:
            raise ConfigException(f"Invalid YAML file:{e}") from None

    return raw_config



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import datetime
import logging
from collections.abc import Iterable, Sequence
from typing import cast
from uuid import UUID

from odc.geo import CRS, Geometry
from typing_extensions import override

from datacube.index.abstract import AbstractIndex, AbstractProductResource
from datacube.index.fields import as_expression
from datacube.model import Product
from datacube.model._base import QueryDict, QueryField
from datacube.utils import _readable_offset, changes, jsonify_document
from datacube.utils.changes import (
    AllowPolicy,
    Change,
    Offset,
    check_doc_unchanged,
    classify_changes,
    get_doc_changes,
)
from datacube.utils.documents import JsonDict, metadata_subset

_LOG: logging.Logger = logging.getLogger(__name__)


class ProductResource(AbstractProductResource):
    def __init__(self, index: AbstractIndex) -> None:
        from datacube.index.memory.index import Index

        self._index: Index = cast(Index, index)
        self.by_id: dict[int, Product] = {}
        self.by_name: dict[str, Product] = {}
        self.next_id = 1

    @override
    def add(self, product: Product, allow_table_lock: bool = False) -> Product | None:
        Product.validate(product.definition)  # type: ignore[attr-defined]
        existing = self.get_by_name(product.name)
        if existing:
            _LOG.warning(
                f"Product {product.name} is already in the database, checking for differences"
            )
            check_doc_unchanged(
                existing.definition,
                jsonify_document(product.definition),
                f"Metadata Type {product.name}",
            )
        else:
            mdt = self._index.metadata_types.get_by_name(product.metadata_type.name)
            if mdt is None:
                _LOG.warning(
                    f'Adding metadata_type "{product.metadata_type.name}" as it doesn\'t exist'
                )
                product.metadata_type = self._index.metadata_types.add(
                    product.metadata_type, allow_table_lock=allow_table_lock
                )
            clone = self.clone(product)
            clone.id = self.next_id
            self.next_id += 1
            self.by_id[clone.id] = clone
            self.by_name[clone.name] = clone
        return self.get_by_name(product.name)

    @override
    def can_update(
        self,
        product: Product,
        allow_unsafe_updates: bool = False,
        allow_table_lock: bool = False,
    ) -> tuple[bool, Iterable[Change], Iterable[Change]]:
        Product.validate(product.definition)  # type: ignore[attr-defined]

        existing = self.get_by_name(product.name)
        if not existing:
            raise ValueError(
                f"Unknown product {product.name}, cannot update - add first"
            )

        updates_allowed: dict[Offset, AllowPolicy] = {
            ("description",): changes.allow_any,
            ("license",): changes.allow_any,
            ("metadata_type",): changes.allow_any,
            # You can safely make the match rules looser but not tighter.
            # Tightening them could exclude datasets already matched to the product.
            # (which would make search results wrong)
            ("metadata",): changes.allow_truncation,
            # Some old storage fields should not be in the product definition any more: allow removal.
            ("storage", "chunking"): changes.allow_removal,
            ("storage", "driver"): changes.allow_removal,
            ("storage", "dimension_order"): changes.allow_removal,
        }
        doc_changes = get_doc_changes(
            existing.definition, jsonify_document(product.definition)
        )
        good_changes, bad_changes = classify_changes(doc_changes, updates_allowed)

        for offset, old_val, new_val in good_changes:
            _LOG.info(
                "Safe change in %s from %r to %r",
                _readable_offset(offset),
                old_val,
                new_val,
            )

        for offset, old_val, new_val in bad_changes:
            _LOG.warning(
                "Unsafe change in %s from %r to %r",
                _readable_offset(offset),
                old_val,
                new_val,
            )

        return (
            (allow_unsafe_updates or not bad_changes),
            good_changes,
            bad_changes,
        )

    @override
    def update(
        self,
        product: Product,
        allow_unsafe_updates: bool = False,
        allow_table_lock: bool = False,
    ) -> Product | None:
        can_update, safe_changes, unsafe_changes = self.can_update(
            product, allow_unsafe_updates
        )

        if not safe_changes and not unsafe_changes:
            _LOG.warning(f"No changes detected for product {product.name}")
            return self.get_by_name(product.name)

        if not can_update:
            errs = ", ".join(
                _readable_offset(offset) for offset, _, _ in unsafe_changes
            )
            raise ValueError(f"Unsafe changes in {product.name}: {errs}")

        existing = cast(Product, self.get_by_name(product.name))
        if product.metadata_type.name != existing.metadata_type.name:
            raise ValueError(
                "Unsafe change: cannot (currently) switch metadata types for a product"
            )

        _LOG.info(f"Updating product {product.name}")

        persisted = self.clone(product)
        persisted.id = cast(int, existing.id)
        self.by_id[persisted.id] = persisted
        self.by_name[persisted.name] = persisted
        return self.get_by_name(product.name)

    @override
    def delete(
        self, products: Iterable[Product], allow_delete_active: bool = False
    ) -> Sequence[Product]:
        deleted = []
        for product in products:
            datasets = self._index.datasets.search_returning(
                ("id",), archived=None, product=product.name
            )
            if datasets:
                purged = self._index.datasets.purge(
                    [ds.id for ds in datasets],  # type: ignore[attr-defined]
                    allow_delete_active,
                )
                if len(purged) != len(list(datasets)):
                    _LOG.warning(
                        f"Product {product.name} cannot be deleted because it has active datasets."
                    )
                    continue
            if product.id is not None:
                del self.by_id[product.id]
            del self.by_name[product.name]
            deleted.append(product)
        return deleted

    @override
    def get_unsafe(self, id_: int) -> Product:
        return self.clone(self.by_id[id_])

    @override
    def get_by_name_unsafe(self, name: str) -> Product:
        return self.clone(self.by_name[name])

    @override
    def search_robust(self, **query: QueryField) -> Iterable[tuple[Product, QueryDict]]:
        def listify(v):
            if isinstance(v, tuple):
                return []
            elif isinstance(v, list):
                return v
            else:
                return [v]

        for prod in self.get_all():
            unmatched = query.copy()
            # Skip non-matched if user specified specific products/metadata_types
            if prod.name not in listify(unmatched.pop("product", prod.name)):
                continue
            if prod.metadata_type.name not in listify(
                unmatched.pop("metadata_type", prod.metadata_type.name)
            ):
                continue
            # Check that all search keys match this product
            for key, value in list(unmatched.items()):
                field = prod.metadata_type.dataset_fields.get(key)
                if not field:
                    # Product doesn't have this field - can't match
                    break
                if not hasattr(field, "extract"):
                    # non-document/native field (??)
                    continue
                if field.extract(prod.metadata_doc) is None:
                    # Product has the field, but not defined in the type doc, so unmatchable
                    continue
                expr = as_expression(field, value)
                if expr.evaluate(prod.metadata_doc):
                    # matches
                    unmatched.pop(key)
                else:
                    # Doesn't match, skip to next product
                    break
            else:
                yield prod, unmatched

    @override
    def search_by_metadata(self, metadata: JsonDict) -> Iterable[Product]:
        norm_meta = {"properties": metadata}
        for prod in self.get_all():
            if metadata_subset(norm_meta, prod.metadata_doc):
                yield prod

    @override
    def get_all(self) -> Iterable[Product]:
        return (self.clone(prod) for prod in self.by_id.values())

    def clone(self, orig: Product) -> Product:
        return Product(
            self._index.metadata_types.clone(orig.metadata_type),
            jsonify_document(orig.definition),
            id_=orig.id,
        )

    @override
    def spatial_extent(
        self, product: str | Product, crs: CRS = CRS("EPSG:4326")
    ) -> Geometry | None:
        return None

    @override
    def temporal_extent(
        self, product: str | Product
    ) -> tuple[datetime.datetime, datetime.datetime]:
        if isinstance(product, str):
            product = self._index.products.get_by_name_unsafe(product)
        ids: Iterable[UUID] = self._index.datasets._by_product.get(product.name, [])
        if len(list(ids)) == 0:
            raise RuntimeError(
                "Product has no datasets and therefore no temporal extent"
            )
        return self._index.datasets.temporal_extent(ids)

    @override
    def most_recent_change(self, product: str | Product) -> datetime.datetime | None:
        raise NotImplementedError(
            "product most recent change is not currently supported by the memory index driver."
        )



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""Add the 'updated' column

Revision ID: 18afe14f2816
Revises: 48fb71ece268
Create Date: 2024-08-08 04:16:19.168213

"""
from typing import Sequence, Union

from alembic import op
from sqlalchemy import Column, DateTime, inspect
from sqlalchemy.sql import func


# revision identifiers, used by Alembic.
revision: str = '18afe14f2816'
down_revision: Union[str, None] = '48fb71ece268'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def column_exists(table_name, column_name):
    bind = op.get_context().bind
    insp = inspect(bind)
    columns = insp.get_columns(table_name, "odc")
    return any(c["name"] == column_name for c in columns)


def upgrade() -> None:
    if not column_exists("dataset", "updated"):
        op.add_column("dataset", Column("updated", DateTime(timezone=True), server_default=func.now(),
                                        nullable=False, index=True, comment="when last updated"), schema="odc")
    else:
        op.alter_column("dataset", "updated", schema="odc", type_=DateTime(timezone=True),
                        server_default=func.now(), nullable=False, comment="when last updated")  # type: ignore[arg-type] # noqa: E501
    op.create_index("ix_dataset_updated", "dataset", ["updated"], schema="odc", if_not_exists=True)

    if not column_exists("metadata_type", "updated"):
        op.add_column("metadata_type", Column("updated", DateTime(timezone=True), server_default=func.now(),
                                              nullable=False, comment="when last updated"), schema="odc")
    else:
        op.alter_column("metadata_type", "updated", schema="odc", type_=DateTime(timezone=True),
                        server_default=func.now(), nullable=False, comment="when last updated")  # type: ignore[arg-type] # noqa: E501

    if not column_exists("product", "updated"):
        op.add_column("product", Column("updated", DateTime(timezone=True), server_default=func.now(),
                                        nullable=False, comment="when last updated"), schema="odc")
    else:
        op.alter_column("product", "updated", schema="odc", type_=DateTime(timezone=True),
                        server_default=func.now(), nullable=False, comment="when last updated")  # type: ignore[arg-type] # noqa: E501


def downgrade() -> None:
    op.drop_column("dataset", "updated", schema="odc")
    op.drop_column("metadata_type", "updated", schema="odc")
    op.drop_column("product", "updated", schema="odc")

    op.drop_index("ix_dataset_updated", schema="odc", if_exists=True)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from typing import cast

from odc.geo.crs import CRS
from odc.geo.geom import Geometry, box

from datacube.model import QueryDict, QueryField, Range
from datacube.utils.documents import JsonDict

H_SPATIAL_KEYS = ("lon", "longitude", "x")
V_SPATIAL_KEYS = ("lat", "latitude", "y")
COORDS_SPATIAL_KEYS: tuple[str, ...] = H_SPATIAL_KEYS + V_SPATIAL_KEYS

CRS_SPATIAL_KEYS = ("crs", "coordinate_reference_system")

# All of the above
NON_GEOPOLYGON_SPATIAL_KEYS: tuple[str, ...] = COORDS_SPATIAL_KEYS + CRS_SPATIAL_KEYS

# All of the above plus geopolygon
SPATIAL_KEYS: tuple[str, ...] = NON_GEOPOLYGON_SPATIAL_KEYS + ("geopolygon",)


def strip_all_spatial_fields_from_query(q: QueryDict) -> QueryDict:
    return {k: v for k, v in q.items() if k not in SPATIAL_KEYS}


def extract_geom_from_query(**q: QueryField) -> Geometry | None:
    """
    Utility method for index drivers supporting spatial indexes.

    Extract a Geometry from a dataset query.  Backwards compatible with old lat/lon style queries.

    :param q: A query dictionary
    :return: A polygon or multipolygon type Geometry.  None if no spatial query clauses.
    """
    geom: Geometry | None = None
    polygon_candidate = q.get("geopolygon")
    if polygon_candidate is not None:
        # New geometry-style spatial query
        geom_term = cast(JsonDict | Geometry, polygon_candidate)
        try:
            geom = Geometry(geom_term)
        except ValueError:
            # Can't convert to single Geometry. If it is an iterable of Geometries, return the union
            for term in geom_term:
                geom = Geometry(term) if geom is None else geom.union(Geometry(term))
        for spatial_key in NON_GEOPOLYGON_SPATIAL_KEYS:
            if spatial_key in q:
                raise ValueError(
                    f"Cannot specify spatial key {spatial_key} AND geopolygon in the same query"
                )
        assert geom and geom.crs
    else:
        # Old lat/lon--style spatial query (or no spatial query)
        # TODO: latitude/longitude/x/y aliases for lat/lon
        #       Also some stuff is precalced at the api.core.Datacube level.
        #       THAT needs to offload to index driver when it can.
        lon = lat = None
        for coord in H_SPATIAL_KEYS:
            if coord in q:
                if lon is not None:
                    raise ValueError(
                        "Multiple horizontal coordinate ranges supplied: use only one of x, lon, longitude"
                    )
                lon = q.get(coord)
        for coord in V_SPATIAL_KEYS:
            if coord in q:
                if lat is not None:
                    raise ValueError(
                        "Multiple vertical coordinate ranges supplied: use only one of y, lat, latitude"
                    )
                lat = q.get(coord)
        crs_in = None
        for coord in CRS_SPATIAL_KEYS:
            if coord in q:
                if crs_in is not None:
                    raise ValueError("CRS is supplied twice")
                crs_in = q.get(coord)
        crs = CRS("epsg:4326") if crs_in is None else CRS(crs_in)
        if lat is None and lon is None:
            # No spatial query
            return None

        # Old lat/lon--style spatial query
        # Normalise input to numeric ranges.
        delta = 0.000001
        if lat is None:
            lat = Range(begin=-90, end=90)
        elif isinstance(lat, int | float):
            lat = Range(lat - delta, lat + delta)
        else:
            # Treat as tuple
            begin, end = cast(tuple[int | float, int | float], lat)
            lat = Range(begin, end)

        if lon is None:
            lon = Range(begin=-180, end=180)
        elif isinstance(lon, int | float):
            lon = Range(lon - delta, lon + delta)
        else:
            # Treat as tuple
            begin, end = cast(tuple[int | float, int | float], lon)
            lon = Range(begin, end)
        geom = box(lon.begin, lat.begin, lon.end, lat.end, crs=crs)
    return geom



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Module
"""

from sqlalchemy.dialects.postgresql import Range as PgRange

from datacube.drivers.postgres._fields import (
    EqualsExpression,
    NumericRangeDocField,
    RangeBetweenExpression,
    SimpleDocField,
)
from datacube.index.fields import to_expressions
from datacube.model import Range


def test_build_query_expressions() -> None:
    _sat_field = SimpleDocField("platform", "platform description", None, False)
    _sens_field = SimpleDocField("instrument", "instrument description", None, False)
    _lat_field = NumericRangeDocField("lat", "latitude range", None, False)
    _fields = {"platform": _sat_field, "instrument": _sens_field, "lat": _lat_field}

    assert [EqualsExpression(_sat_field, "LANDSAT_8")] == to_expressions(
        _fields.get, platform="LANDSAT_8"
    )
    assert [
        RangeBetweenExpression(_lat_field, 4, 23.0, _range_class=PgRange)
    ] == to_expressions(_fields.get, lat=Range(4, 23))



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import logging
import math

import pytest
import toolz
import yaml

from datacube.index import Index
from datacube.index.hl import Doc2Dataset
from datacube.model import MetadataType
from datacube.scripts.dataset import _resolve_uri
from datacube.testutils import (
    dataset_maker,
    gen_dataset_test_dag,
    load_dataset_definition,
    write_files,
)
from datacube.utils import SimpleDocNav

logger = logging.getLogger(__name__)


def check_skip_lineage_test(clirunner, index) -> None:
    ds = SimpleDocNav(gen_dataset_test_dag(11, force_tree=True))

    prefix = write_files({"agdc-metadata.yml": yaml.safe_dump(ds.doc)})

    clirunner(["dataset", "add", "--ignore-lineage", "--product", "A", str(prefix)])

    ds_ = index.datasets.get(ds.id, include_sources=True)
    assert ds_ is not None
    assert ds_.id == ds.id
    assert ds_.sources == {}

    assert index.datasets.get(ds.sources["ab"].id) is None
    assert index.datasets.get(ds.sources["ac"].id) is None
    assert index.datasets.get(ds.sources["ae"].id) is None
    assert index.datasets.get(ds.sources["ac"].sources["cd"].id) is None


def check_no_product_match(clirunner, index) -> None:
    ds = SimpleDocNav(gen_dataset_test_dag(22, force_tree=True))

    prefix = write_files({"agdc-metadata.yml": yaml.safe_dump(ds.doc)})

    r = clirunner(["dataset", "add", "--product", "A", str(prefix)])
    assert "ERROR" in r.output
    assert "Dataset metadata did not match product signature" in r.output

    r = clirunner(["dataset", "add", "--product", "A", "--product", "B", str(prefix)])
    assert "ERROR" in r.output
    assert "No matching Product found for dataset" in r.output

    ds_ = index.datasets.get(ds.id, include_sources=True)
    assert ds_ is None

    # Ignore lineage but fail to match main dataset
    r = clirunner(["dataset", "add", "--product", "B", "--ignore-lineage", str(prefix)])

    assert "ERROR" in r.output
    assert "Dataset metadata did not match product signature" in r.output
    assert index.datasets.has(ds.id) is False


def check_with_existing_lineage(clirunner, index) -> None:
    """
      A -> B
      |    |
      |    v
      +--> C -> D
      |
      +--> E

    Add nodes BCE(D) with auto-matching, then add node A with product restricted to A only.
    """
    ds = SimpleDocNav(gen_dataset_test_dag(33, force_tree=True))

    child_docs = [ds.sources[x].doc for x in ("ab", "ac", "ae")]

    prefix = write_files(
        {
            "lineage.yml": yaml.safe_dump_all(child_docs),
            "main.yml": yaml.safe_dump(ds.doc),
        }
    )

    clirunner(["dataset", "add", str(prefix / "lineage.yml")])
    assert index.datasets.get(ds.sources["ae"].id) is not None
    assert index.datasets.get(ds.sources["ab"].id) is not None
    assert index.datasets.get(ds.sources["ac"].id) is not None

    clirunner(
        [
            "dataset",
            "add",
            "--no-auto-add-lineage",
            "--product",
            "A",
            str(prefix / "main.yml"),
        ]
    )

    assert index.datasets.get(ds.id) is not None


def check_inconsistent_lineage(clirunner, index) -> None:
    """
      A -> B
      |    |
      |    v
      +--> C -> D
      |
      +--> E

    Add node E,
    then try adding A with modified E in the lineage, should fail to add ABCD
    """
    ds = SimpleDocNav(gen_dataset_test_dag(1313, force_tree=True))

    child_docs = [ds.sources[x].doc for x in ("ae",)]
    modified_doc = toolz.assoc_in(
        ds.doc, ["lineage", "source_datasets", "ae", "label"], "modified"
    )

    prefix = write_files(
        {
            "lineage.yml": yaml.safe_dump_all(child_docs),
            "main.yml": yaml.safe_dump(modified_doc),
        }
    )

    clirunner(["dataset", "add", str(prefix / "lineage.yml")])
    assert index.datasets.get(ds.sources["ae"].id) is not None

    r = clirunner(["dataset", "add", str(prefix / "main.yml")])

    assert "ERROR Inconsistent lineage dataset" in r.output

    assert index.datasets.has(ds.id) is False
    assert index.datasets.has(ds.sources["ab"].id) is False
    assert index.datasets.has(ds.sources["ac"].id) is False
    assert index.datasets.has(ds.sources["ac"].sources["cd"].id) is False

    # now again but skipping verification check
    _ = clirunner(["dataset", "add", "--no-verify-lineage", str(prefix / "main.yml")])

    assert index.datasets.has(ds.id)
    assert index.datasets.has(ds.sources["ab"].id)
    assert index.datasets.has(ds.sources["ac"].id)
    assert index.datasets.has(ds.sources["ac"].sources["cd"].id)


def check_missing_lineage(clirunner, index) -> None:
    """
      A -> B
      |    |
      |    v
      +--> C -> D
      |
      +--> E

    Use --no-auto-add-lineage
    """
    ds = SimpleDocNav(gen_dataset_test_dag(44, force_tree=True))
    child_docs = [ds.sources[x].doc for x in ("ae", "ab", "ac")]

    prefix = write_files(
        {
            "lineage.yml": yaml.safe_dump_all(child_docs),
            "main.yml": yaml.safe_dump(ds.doc),
        }
    )

    r = clirunner(["dataset", "add", "--no-auto-add-lineage", str(prefix / "main.yml")])

    assert "ERROR Following lineage datasets are missing" in r.output
    assert index.datasets.has(ds.id) is False

    # now add lineage and try again
    clirunner(["dataset", "add", str(prefix / "lineage.yml")])
    assert index.datasets.has(ds.sources["ae"].id)
    _ = clirunner(["dataset", "add", "--no-auto-add-lineage", str(prefix / "main.yml")])

    assert index.datasets.has(ds.id)


def check_missing_metadata_doc(clirunner) -> None:
    prefix = write_files({"im.tiff": ""})
    r = clirunner(["dataset", "add", str(prefix / "im.tiff")])
    assert "ERROR No supported metadata docs found for dataset" in r.output


def check_bad_yaml(clirunner, index) -> None:
    prefix = write_files({"broken.yml": '"'})
    r = clirunner(["dataset", "add", str(prefix / "broken.yml")])
    assert "ERROR Failed reading documents from " in r.output


def test_dataset_add_no_id(index: Index, eo3_ls8_dataset3_doc, ls8_eo3_product) -> None:
    doc, uri = eo3_ls8_dataset3_doc
    del doc["id"]

    # Check .hl.Doc2Dataset
    doc2ds = Doc2Dataset(index)
    _ds, _err = doc2ds(doc, uri)
    assert _err == "No id defined in dataset doc"


def test_dataset_add_not_eo3(index: Index, ls8_eo3_product, eo3_wo_dataset_doc) -> None:
    from datacube.model.utils import BadMatch

    doc2ds = Doc2Dataset(index)
    _ds, _err = doc2ds(*eo3_wo_dataset_doc)
    assert isinstance(_err, BadMatch)


@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_dataset_eo3_no_schema(
    dataset_add_configs, index_empty, clirunner, caplog
) -> None:
    p = dataset_add_configs
    index = index_empty
    tmp = load_dataset_definition(p.datasets_eo3)
    assert tmp is not None
    ds = tmp.doc

    clirunner(["metadata", "add", p.metadata])
    clirunner(["product", "add", p.products])

    # no warnings if eo3 dataset has $schema
    doc2ds = Doc2Dataset(index)
    doc2ds(ds, "file:///something")
    warnings = [
        record for record in caplog.records if record.levelno == logging.WARNING
    ]
    assert len(warnings) == 0

    # warn if eo3 metadata type but no $schema
    del ds["$schema"]
    doc2ds(ds, "file:///something")
    warnings = [
        record for record in caplog.records if record.levelno == logging.WARNING
    ]
    assert len(warnings) == 1
    assert "will not be recognised as an eo3 dataset" in warnings[0].msg


# Current formulation of this test relies on non-EO3 test data
@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_dataset_add(dataset_add_configs, index_empty, clirunner) -> None:
    p = dataset_add_configs
    index = index_empty
    r = clirunner(["dataset", "add", p.datasets], expect_success=False)
    assert r.exit_code != 0
    assert "Found no matching products" in r.output

    clirunner(["metadata", "add", p.metadata])
    clirunner(["product", "add", p.products])
    clirunner(["dataset", "add", p.datasets])
    clirunner(["dataset", "add", p.datasets_bad1])
    clirunner(["dataset", "add", p.datasets_eo3])

    ds = load_dataset_definition(p.datasets)
    assert ds is not None
    ds_bad1 = load_dataset_definition(p.datasets_bad1)
    assert ds_bad1 is not None

    # Check .hl.Doc2Dataset
    doc2ds = Doc2Dataset(index)
    _ds, _err = doc2ds(ds.doc, "file:///something")
    assert _err is None
    assert _ds is not None
    assert _ds.id == ds.id
    assert _ds.metadata_doc == ds.doc

    # Check dataset search
    r = clirunner(["dataset", "search"])
    assert str(ds.id) in r.output
    assert str(ds_bad1.id) not in r.output
    assert str(ds.sources["ab"].id) in r.output
    assert str(ds.sources["ac"].sources["cd"].id) in r.output

    r = clirunner(["dataset", "info", "-f", "csv", str(ds.id)])
    assert str(ds.id) in r.output

    r = clirunner(["dataset", "info", "-f", "yaml", "--show-sources", str(ds.id)])
    assert str(ds.sources["ae"].id) in r.output

    r = clirunner(
        ["dataset", "info", "-f", "yaml", "--show-derived", str(ds.sources["ae"].id)]
    )
    assert str(ds.id) in r.output

    ds_ = SimpleDocNav(gen_dataset_test_dag(1, force_tree=True))
    assert ds_.id == ds.id

    x = index.datasets.get(ds.id, include_sources=True)
    assert x is not None
    assert x.sources["ab"].id == ds.sources["ab"].id
    assert x.sources["ac"].sources["cd"].id == ds.sources["ac"].sources["cd"].id

    check_skip_lineage_test(clirunner, index)
    check_no_product_match(clirunner, index)
    check_with_existing_lineage(clirunner, index)
    check_inconsistent_lineage(clirunner, index)
    check_missing_metadata_doc(clirunner)
    check_missing_lineage(clirunner, index)
    check_bad_yaml(clirunner, index)

    # check --product=nosuchproduct
    r = clirunner(
        ["dataset", "add", "--product", "nosuchproduct", p.datasets],
        expect_success=False,
    )

    assert "ERROR Supplied product name" in r.output
    assert r.exit_code != 0

    # test dataset add eo3
    r = clirunner(["dataset", "add", p.datasets_eo3])
    assert r.exit_code == 0

    ds_eo3 = load_dataset_definition(p.datasets_eo3)
    assert ds_eo3 is not None
    assert ds_eo3.location is not None

    _ds = index.datasets.get(ds_eo3.id, include_sources=True)
    assert _ds is not None
    assert sorted(_ds.sources) == ["a", "bc1", "bc2"]
    assert _ds.crs == "EPSG:3857"
    assert _ds.extent is not None
    assert _ds.extent.crs == _ds.crs
    assert _ds.uri == ds_eo3.location
    assert "location" not in _ds.metadata_doc


# Current formulation of this test relies on non-EO3 test data
@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_dataset_add_ambiguous_products(
    dataset_add_configs, index_empty, clirunner
) -> None:
    p = dataset_add_configs
    index = index_empty

    dss = [
        SimpleDocNav(dataset_maker(i)("A", product_type="eo", flag_a="a", flag_b="b"))
        for i in [1, 2]
    ]

    prefix = write_files(
        {
            "products.yml": """
name: A
description: test product A
metadata_type: minimal
metadata:
    product_type: eo
    flag_a: a

---
name: B
description: test product B
metadata_type: minimal
metadata:
    product_type: eo
    flag_b: b
    """,
            "dataset1.yml": yaml.safe_dump(dss[0].doc),
            "dataset2.yml": yaml.safe_dump(dss[1].doc),
        }
    )

    clirunner(["metadata", "add", p.metadata])
    clirunner(["product", "add", str(prefix / "products.yml")])

    pp = list(index.products.get_all())
    assert len(pp) == 2

    for ds, i in zip(dss, (1, 2)):
        r = clirunner(["dataset", "add", str(prefix / f"dataset{i}.yml")])
        assert "ERROR Auto match failed" in r.output
        assert "matches several products" in r.output
        assert index.datasets.has(ds.id) is False

    # check that forcing product works
    ds, fname = dss[0], "dataset1.yml"
    _ = clirunner(["dataset", "add", "--product", "A", str(prefix / fname)])

    assert index.datasets.has(ds.id) is True

    # check that forcing via exclude works
    ds, fname = dss[1], "dataset2.yml"
    _ = clirunner(["dataset", "add", "--exclude-product", "B", str(prefix / fname)])

    assert index.datasets.has(ds.id) is True


# Current formulation of this test relies on non-EO3 test data
@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_dataset_add_with_nans(dataset_add_configs, index_empty, clirunner) -> None:
    p = dataset_add_configs
    index = index_empty

    clirunner(["metadata", "add", p.metadata])
    clirunner(["product", "add", p.products])

    mk = dataset_maker(0)

    c = mk(
        "C",
        product_type="C",
        val_is_nan=math.nan,
        val_is_inf=math.inf,
        val_is_neginf=-math.inf,
    )

    b = mk("B", sources={"bc": c}, product_type="B")
    a = mk("A", sources={"ac": c}, product_type="A")

    prefix = write_files(
        {
            "dataset.yml": yaml.safe_dump_all([a, b]),
        }
    )

    r = clirunner(
        [
            "dataset",
            "add",
            "--auto-add-lineage",
            "--verify-lineage",
            str(prefix / "dataset.yml"),
        ]
    )

    assert "ERROR" not in r.output

    a, b, c = [SimpleDocNav(v) for v in (a, b, c)]

    assert index.datasets.bulk_has([a.id, b.id, c.id]) == [True, True, True]

    c_doc = index.datasets.get(c.id).metadata_doc

    assert c_doc["val_is_nan"] == "NaN"
    assert c_doc["val_is_inf"] == "Infinity"
    assert c_doc["val_is_neginf"] == "-Infinity"


# Current formulation of this test relies on non-EO3 test data
@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_dataset_add_inconsistent_measurements(
    dataset_add_configs, index_empty, clirunner
) -> None:
    p = dataset_add_configs
    index = index_empty
    mk = dataset_maker(0)

    # not set, empty, subset, full set, super-set
    ds1 = SimpleDocNav(
        mk(
            "A",
            product_type="eo",
        )
    )
    ds2 = SimpleDocNav(mk("B", product_type="eo", measurements={}))
    ds3 = SimpleDocNav(mk("C", product_type="eo", measurements={"red": {}}))
    ds4 = SimpleDocNav(
        mk(
            "D",
            product_type="eo",
            measurements={
                "red": {},
                "green": {},
            },
        )
    )
    ds5 = SimpleDocNav(
        mk(
            "E",
            product_type="eo",
            measurements={
                "red": {},
                "green": {},
                "extra": {},
            },
        )
    )

    dss = (ds1, ds2, ds3, ds4, ds5)
    docs = [ds.doc for ds in dss]

    prefix = write_files(
        {
            "products.yml": """
name: eo
description: test product
metadata_type: with_measurements
metadata:
    product_type: eo

measurements:
    - name: red
      dtype: int16
      nodata: -999
      units: '1'

    - name: green
      dtype: int16
      nodata: -999
      units: '1'
    """,
            "dataset.yml": yaml.safe_dump_all(docs),
        }
    )

    clirunner(["metadata", "add", p.metadata])
    _ = clirunner(["product", "add", str(prefix / "products.yml")])

    pp = list(index.products.get_all())
    assert len(pp) == 1

    r = clirunner(["dataset", "add", str(prefix / "dataset.yml")])
    print(r.output)

    r = clirunner(["dataset", "search", "-f", "csv"])
    assert str(ds1.id) not in r.output
    assert str(ds2.id) not in r.output
    assert str(ds3.id) not in r.output
    assert str(ds4.id) in r.output
    assert str(ds5.id) in r.output


def dataset_archive_prep(dataset_add_configs, index_empty, clirunner):
    p = dataset_add_configs
    index = index_empty

    clirunner(["metadata", "add", p.metadata])
    clirunner(["product", "add", p.products])
    clirunner(["dataset", "add", p.datasets])

    ds = load_dataset_definition(p.datasets)

    assert index.datasets.has(ds.id) is True

    return p, index, ds


# Current formulation of this test relies on non-EO3 test data
@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_dataset_archive_dry_run(dataset_add_configs, index_empty, clirunner) -> None:
    p, index, ds = dataset_archive_prep(dataset_add_configs, index_empty, clirunner)

    non_existent_uuid = "00000000-1036-5607-a62f-fde5e3fec985"

    # Single valid UUID is detected and not archived
    single_valid_uuid = clirunner(["dataset", "archive", "--dry-run", str(ds.id)])
    assert str(ds.id) in single_valid_uuid.output
    assert index.datasets.has(ds.id) is True

    # Single invalid UUID is detected
    single_invalid_uuid = clirunner(
        ["dataset", "archive", "--dry-run", non_existent_uuid], expect_success=False
    )
    assert single_invalid_uuid.exit_code is -1
    assert non_existent_uuid in single_invalid_uuid.output
    assert index.datasets.has(ds.id) is True

    # Valid and invalid UUIDs
    # The single invalid UUID should halt all operations
    valid_and_invalid_uuid = clirunner(
        ["dataset", "archive", "--dry-run", str(ds.id), non_existent_uuid],
        expect_success=False,
    )
    assert non_existent_uuid in valid_and_invalid_uuid.output
    assert single_invalid_uuid.exit_code is -1
    assert index.datasets.has(ds.id) is True

    valid_and_invalid_uuid = clirunner(
        [
            "dataset",
            "archive",
            "--dry-run",
            "--archive-derived",
            str(ds.id),
            non_existent_uuid,
        ],
        expect_success=False,
    )
    assert single_invalid_uuid.exit_code is -1
    assert non_existent_uuid in valid_and_invalid_uuid.output
    assert index.datasets.has(ds.id) is True

    # Multiple Valid UUIDs
    # Not archived in the database and are shown in output
    multiple_valid_uuid = clirunner(
        [
            "dataset",
            "archive",
            "--dry-run",
            str(ds.sources["ae"].id),
            str(ds.sources["ab"].id),
        ]
    )
    assert str(ds.sources["ae"].id) in multiple_valid_uuid.output
    assert str(ds.sources["ab"].id) in multiple_valid_uuid.output
    assert index.datasets.has(ds.sources["ae"].id) is True
    assert index.datasets.has(ds.sources["ab"].id) is True

    archive_derived = clirunner(
        [
            "dataset",
            "archive",
            "--dry-run",
            "--archive-derived",
            str(ds.sources["ae"].id),
            str(ds.sources["ab"].id),
        ]
    )
    assert str(ds.id) in archive_derived.output
    assert str(ds.sources["ae"].id) in archive_derived.output
    assert index.datasets.has(ds.id) is True


# Current formulation of this test relies on non-EO3 test data
@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_dataset_archive_restore_invalid(
    dataset_add_configs, index_empty, clirunner
) -> None:
    p, index, ds = dataset_archive_prep(dataset_add_configs, index_empty, clirunner)

    non_existent_uuid = "00000000-1036-5607-a62f-fde5e3fec985"

    # With non-existent uuid, operations should halt.
    _ = clirunner(
        ["dataset", "archive", str(ds.id), non_existent_uuid], expect_success=False
    )
    r = clirunner(["dataset", "info", str(ds.id)])
    assert "status: archived" not in r.output
    assert index.datasets.has(ds.id) is True

    # With non-existent uuid, operations should halt.
    d_id = ds.sources["ac"].sources["cd"].id
    _ = clirunner(
        ["dataset", "archive", "--archive-derived", str(d_id), non_existent_uuid],
        expect_success=False,
    )
    r = clirunner(
        [
            "dataset",
            "info",
            str(ds.id),
            str(ds.sources["ab"].id),
            str(ds.sources["ac"].id),
        ]
    )
    assert "status: active" in r.output
    assert "status: archived" not in r.output
    assert index.datasets.has(ds.id) is True


# Current formulation of this test relies on non-EO3 test data
@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_dataset_archive_restore(dataset_add_configs, index_empty, clirunner) -> None:
    p, index, ds = dataset_archive_prep(dataset_add_configs, index_empty, clirunner)

    # Run for real
    _ = clirunner(["dataset", "archive", str(ds.id)])
    r = clirunner(["dataset", "info", str(ds.id)])
    assert "status: archived" in r.output

    # restore dry run
    _ = clirunner(["dataset", "restore", "--dry-run", str(ds.id)])
    r = clirunner(["dataset", "info", str(ds.id)])
    assert "status: archived" in r.output

    # restore for real
    _ = clirunner(["dataset", "restore", str(ds.id)])
    r = clirunner(["dataset", "info", str(ds.id)])
    assert "status: active" in r.output

    # archive derived
    d_id = ds.sources["ac"].sources["cd"].id
    _ = clirunner(["dataset", "archive", "--archive-derived", str(d_id)])
    r = clirunner(
        [
            "dataset",
            "info",
            str(ds.id),
            str(ds.sources["ab"].id),
            str(ds.sources["ac"].id),
        ]
    )
    assert "status: active" not in r.output
    assert "status: archived" in r.output

    # restore derived
    _ = clirunner(["dataset", "restore", "--restore-derived", str(d_id)])
    r = clirunner(
        [
            "dataset",
            "info",
            str(ds.id),
            str(ds.sources["ab"].id),
            str(ds.sources["ac"].id),
        ]
    )
    assert "status: active" in r.output
    assert "status: archived" not in r.output


# Current formulation of this test relies on non-EO3 test data
@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_dataset_add_http(
    dataset_add_configs,
    index: Index,
    default_metadata_type: MetadataType,
    httpserver,
    clirunner,
) -> None:
    # pytest-localserver also looks good, it's been around for ages, but httpserver is the new cool
    p = dataset_add_configs

    httpserver.expect_request("/metadata_types.yaml").respond_with_data(
        open(p.metadata).read()
    )
    httpserver.expect_request("/products.yaml").respond_with_data(
        open(p.products).read()
    )
    httpserver.expect_request("/datasets.yaml").respond_with_data(
        open(p.datasets).read()
    )
    # check that the request is served
    #    assert requests.get(httpserver.url_for("/dataset.yaml")).yaml() == {'foo': 'bar'}

    clirunner(["metadata", "add", httpserver.url_for("/metadata_types.yaml")])
    clirunner(["product", "add", httpserver.url_for("/products.yaml")])
    # clirunner(['dataset', 'add', p.datasets])
    clirunner(["dataset", "add", httpserver.url_for("/datasets.yaml")])

    ds = load_dataset_definition(p.datasets)
    assert ds is not None
    assert index.datasets.has(ds.id)


def xtest_dataset_add_fails(clirunner, index) -> None:
    result = clirunner(["dataset", "add", "bad_path.yaml"], expect_success=False)
    assert result.exit_code != 0, (
        "Surely not being able to add a dataset when requested should return an error."
    )


def test_resolve_uri() -> None:
    def doc(loc=None):
        return SimpleDocNav(
            {"location": loc, "id": "4d9fd75c-1309-4712-93b5-f0d9c6fdd8ab"}
        )

    uri = "file:///a"
    override = "https://a.com/b.yaml"
    assert _resolve_uri(uri, doc()) is uri
    assert _resolve_uri(uri, doc([])) is uri
    assert _resolve_uri(uri, doc({})) is uri
    assert _resolve_uri(uri, doc(object())) is uri
    assert _resolve_uri(uri, doc(override)) is override
    assert _resolve_uri(uri, doc([override])) is override
    assert _resolve_uri(uri, doc([override, "something"])) is override
    assert _resolve_uri(uri, doc((override, "something"))) is override



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import logging
from collections.abc import Iterable, Mapping
from copy import deepcopy
from typing import Any, cast

from typing_extensions import override

from datacube.index.abstract import (
    AbstractMetadataTypeResource,
    default_metadata_type_docs,
)
from datacube.index.memory._fields import get_dataset_fields
from datacube.model import MetadataType
from datacube.utils import _readable_offset, changes, jsonify_document
from datacube.utils.changes import (
    AllowPolicy,
    Change,
    Offset,
    check_doc_unchanged,
    get_doc_changes,
)
from datacube.utils.documents import JsonDict

_LOG: logging.Logger = logging.getLogger(__name__)


class MetadataTypeResource(AbstractMetadataTypeResource):
    def __init__(self) -> None:
        self.by_id: dict = {}
        self.by_name: dict = {}
        self.next_id = 1
        for doc in default_metadata_type_docs():
            self.add(self.from_doc(doc))

    @override
    def from_doc(self, definition: JsonDict) -> MetadataType:
        MetadataType.validate(definition)  # type: ignore[attr-defined]
        return self._make(definition)

    @override
    def add(
        self, metadata_type: MetadataType, allow_table_lock: bool = False
    ) -> MetadataType:
        MetadataType.validate(metadata_type.definition)  # type: ignore[attr-defined]
        if metadata_type.name in self.by_name:
            # Error unless it's the exact same metadata_type
            _LOG.warning("Metadata Type exists, checking for differences")
            check_doc_unchanged(
                self.by_name[metadata_type.name].definition,
                jsonify_document(metadata_type.definition),
                f"Metadata Type {metadata_type.name}",
            )
        else:
            persisted = self._make(
                jsonify_document(metadata_type.definition), id_=self.next_id
            )
            self.next_id += 1
            self.by_id[persisted.id] = persisted
            self.by_name[persisted.name] = persisted
        return cast(MetadataType, self.get_by_name(metadata_type.name))

    @override
    def can_update(
        self, metadata_type: MetadataType, allow_unsafe_updates: bool = False
    ) -> tuple[bool, Iterable[Change], Iterable[Change]]:
        MetadataType.validate(metadata_type.definition)  # type: ignore[attr-defined]
        existing = self.get_by_name(metadata_type.name)
        if not existing:
            raise ValueError(
                f"Unknown metadata type {metadata_type.name}, cannot update - add first"
            )
        updates_allowed: Mapping[Offset, AllowPolicy] = {
            ("description",): changes.allow_any,
            # You can add new fields safely but not modify existing ones.
            ("dataset",): changes.allow_extension,
            ("dataset", "search_fields"): changes.allow_extension,
        }

        doc_changes = get_doc_changes(
            existing.definition, jsonify_document(metadata_type.definition)
        )
        good_changes, bad_changes = changes.classify_changes(
            doc_changes, updates_allowed
        )
        for offset, old_val, new_val in good_changes:
            _LOG.info(
                "Safe change in %s from %r to %r",
                _readable_offset(offset),
                old_val,
                new_val,
            )

        for offset, old_val, new_val in bad_changes:
            _LOG.warning(
                "Unsafe change in %s from %r to %r",
                _readable_offset(offset),
                old_val,
                new_val,
            )

        return ((allow_unsafe_updates or not bad_changes), good_changes, bad_changes)

    @override
    def update(
        self,
        metadata_type: MetadataType,
        allow_unsafe_updates: bool = False,
        allow_table_lock: bool = False,
    ) -> MetadataType:
        can_update, safe_changes, unsafe_changes = self.can_update(
            metadata_type, allow_unsafe_updates
        )

        if not safe_changes and not unsafe_changes:
            _LOG.warning(f"No changes detected for metadata type {metadata_type.name}")
            return self.get_by_name_unsafe(metadata_type.name)

        if not can_update:
            errs = ", ".join(_readable_offset(change[0]) for change in unsafe_changes)
            raise ValueError(f"Unsafe changes in {metadata_type.name}: {errs}")

        _LOG.info(f"Updating metadata type {metadata_type.name}")

        persisted = self.clone(metadata_type)
        self.by_id[metadata_type.id] = persisted
        self.by_name[metadata_type.name] = persisted
        return persisted

    @override
    def get_unsafe(self, id_: int) -> MetadataType:
        return self.clone(self.by_id[id_])

    @override
    def get_by_name_unsafe(self, name: str) -> MetadataType:
        return self.clone(self.by_name[name])

    @override
    def check_field_indexes(
        self,
        allow_table_lock: bool = False,
        rebuild_views: bool = False,
        rebuild_indexes: bool = False,
    ) -> None:
        # Cannot implement this method without separating index implementation into
        # separate layer from the API Resource implementations.
        pass

    @override
    def get_all(self) -> Iterable[MetadataType]:
        return (self.clone(mdt) for mdt in self.by_id.values())

    @staticmethod
    def _make(definition: Mapping[str, Any], id_=None) -> MetadataType:
        # Note this is an index-specific implementation of get_dataset_fields, not the default
        return MetadataType(
            definition, search_field_extractor=get_dataset_fields, id_=id_
        )

    @staticmethod
    def clone(orig: MetadataType) -> MetadataType:
        return MetadataTypeResource._make(deepcopy(orig.definition), id_=orig.id)
